#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import yaml
import tempfile
import subprocess
import traceback
import datetime
import time
import json
import binascii
import os

from ckan_cloud_operator import kubectl
from ckan_cloud_operator.drivers.kubectl import rbac as kubectl_rbac_driver
from ckan_cloud_operator import logs
from ckan_cloud_operator.drivers.helm import driver as helm_driver
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers.ckan import manager as ckan_manager
from ckan_cloud_operator.providers.db import manager as db_manager
from ckan_cloud_operator.providers.solr import manager as solr_manager
from ckan_cloud_operator.annotations import manager as annotations_manager
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.providers.ckan.constants import INSTANCE_CRD_SINGULAR
from ckan_cloud_operator.routers import manager as routers_manager


def initialize(interactive=False):
    tiller_namespace_name = _get_resource_name()
    helm_driver.init(tiller_namespace_name)


def update(instance_id, instance, force=False, dry_run=False):
    tiller_namespace_name = _get_resource_name()
    logs.debug('Updating helm-based instance deployment',
               instance_id=instance_id, tiller_namespace_name=tiller_namespace_name)
    _create_private_container_registry_secret(instance_id)
    _init_ckan_infra_secret(instance_id, dry_run=dry_run)
    ckan_helm_chart_repo = instance['spec'].get(
        "ckanHelmChartRepo",
        "https://raw.githubusercontent.com/ViderumGlobal/ckan-cloud-helm/master/charts_repository"
    )
    ckan_helm_chart_version = instance['spec'].get("ckanHelmChartVersion", "")
    ckan_helm_release_name = f'ckan-cloud-{instance_id}'
    solr_schema = instance['spec'].get("ckanSolrSchema", "ckan_default")
    solr_host, solr_port = _init_solr(instance_id, solr_schema, dry_run=dry_run,)
    logs.debug(ckan_helm_chart_repo=ckan_helm_chart_repo,
               ckan_helm_chart_version=ckan_helm_chart_version, ckan_helm_release_name=ckan_helm_release_name,
               solr_host=solr_host, solr_port=solr_port)
    instance['spec']['centralizedSolrHost'], instance['spec']['centralizedSolrPort'] = solr_host, solr_port
    if annotations_manager.get_status(instance, 'helm', 'created'):
        logs.info('Updating existing instance')
        values = instance['spec']
    else:
        logs.info('New instance, deploying first with 1 replica and disabled probes and jobs')
        values = {
            **instance['spec'],
            "replicas": 1,
            "nginxReplicas": 1,
            "disableJobs": True,
            "noProbes": True,
            "enableHarvesterNG": False
        }
    _helm_deploy(
        values, tiller_namespace_name, ckan_helm_chart_repo, ckan_helm_chart_version,
        ckan_helm_release_name, instance_id, dry_run=dry_run
    )
    if not dry_run:
        _wait_instance_events(instance_id)
        instance = crds_manager.get(INSTANCE_CRD_SINGULAR, name=instance_id)
        if not annotations_manager.get_status(instance, 'helm', 'created'):
            annotations_manager.set_status(instance, 'helm', 'created')
            _helm_deploy(
                instance['spec'], tiller_namespace_name, ckan_helm_chart_repo, ckan_helm_chart_version,
                ckan_helm_release_name, instance_id
            )
            _scale_down_scale_up(namespace=instance_id, replicas=values.get('replicas', 1))


def _helm_deploy(values, tiller_namespace_name, ckan_helm_chart_repo, ckan_helm_chart_version, ckan_helm_release_name,
                 instance_id, dry_run=False):
    logs.debug(f'Deploying helm chart {ckan_helm_chart_repo} {ckan_helm_chart_version} to release {ckan_helm_release_name} (instance_id={instance_id})')
    with tempfile.NamedTemporaryFile('w') as f:
        yaml.dump(values, f, default_flow_style=False)
        f.flush()
        helm_driver.deploy(tiller_namespace_name, ckan_helm_chart_repo, 'ckan-cloud/ckan', ckan_helm_chart_version,
                           ckan_helm_release_name, f.name, instance_id, dry_run=dry_run)


def delete(instance_id, instance):
    tiller_namespace_name = _get_resource_name()
    ckan_helm_release_name = f'ckan-cloud-{instance_id}'
    errors = []
    try:
        logs.info(f'Deleting helm release {ckan_helm_release_name}')
        helm_driver.delete(tiller_namespace_name, ckan_helm_release_name)
    except Exception as e:
        logs.warning(traceback.format_exc())
        errors.append(f'Failed to delete helm release')
    if kubectl.call(f'delete --wait=false namespace {instance_id}') != 0:
        errors.append(f'Failed to delete namespace')
    assert len(errors) == 0, ', '.join(errors)


def get(instance_id, instance=None):
    image = None
    latest_operator_timestamp, latest_pod_name, latest_pod_status = None, None, None
    item_app_statuses = {}
    ckan_deployment_status = None
    ckan_deployment_ready = None
    ckan_deployment_status_pods = []
    logs.debug('Getting all namespace resources', namespace=instance_id)
    all_resources = kubectl.get('all', namespace=instance_id, required=False)
    num_resource_items = len(all_resources.get('items'))
    logs.debug(num_resource_items=num_resource_items)
    if num_resource_items > 0:
        for item in all_resources['items']:
            item_kind = item['kind']
            try:
                if item_kind in ("Pod", "ReplicaSet"):
                    item_app = item["metadata"]["labels"]["app"]
                elif item_kind in ("Service", "Deployment"):
                    item_app = item["metadata"]["name"]
                else:
                    item_app = None
            except:
                logging.exception('Failed to extract item_app from %r', item)
                item_app = None
            logs.debug(item_kind=item_kind, item_app=item_app)
            if item_app in ["ckan", "jobs-db", "redis", "nginx", "jobs"]:
                app_status = item_app_statuses.setdefault(item_app, {})
            else:
                app_status = item_app_statuses.setdefault("unknown", {})
            item_status = kubectl.get_item_detailed_status(item)
            app_status.setdefault("{}s".format(item_kind.lower()), []).append(item_status)
            if item_app == 'ckan':
                if item_kind == 'Deployment':
                    ckan_deployment_status = item_status
                    ckan_deployment_ready = len(item_status.get('error', [])) == 0
                    logs.debug(ckan_deployment_ready=ckan_deployment_ready)
                elif item_kind == 'Pod':
                    pod = item
                    pod_status = item_status
                    pod_operator_timestamp = pod['metadata']['creationTimestamp']
                    if not latest_operator_timestamp or latest_operator_timestamp < pod_operator_timestamp:
                        latest_operator_timestamp = pod_operator_timestamp
                        latest_pod_name = pod['metadata']['name']
                    for container in ["secrets", "ckan"]:
                        status_code, output = subprocess.getstatusoutput(
                            f'kubectl -n {instance_id} logs {pod["metadata"]["name"]} -c {container}',
                        )
                        container_logs = output if status_code == 0 else None
                        logs.debug(len_container_logs=len(container_logs) if container_logs else 0)
                        if container == 'ckan':
                            pod_status['logs'] = output
                        if container_logs:
                            for logline in container_logs.split("--START_CKAN_CLOUD_LOG--")[1:]:
                                logdata = json.loads(logline.split("--END_CKAN_CLOUD_LOG--")[0])
                                pod_status.setdefault("ckan-cloud-logs", []).append(logdata)
                    if not image:
                        image = pod["spec"]["containers"][0]["image"]
                    else:
                        if image != pod["spec"]["containers"][0]["image"]:
                            ckan_deployment_ready = False
                            image = pod["spec"]["containers"][0]["image"]
                    ckan_deployment_status_pods.append(pod_status)
                    if latest_pod_name == pod_status['name']:
                        latest_pod_status = pod_status
        if not latest_pod_status or len(latest_pod_status.get('errors', [])) > 0 or latest_pod_status['logs'] is None:
            ckan_deployment_ready = False
    else:
        ckan_deployment_ready = False
    return {
        **ckan_deployment_status,
        'ready': ckan_deployment_ready,
        'pods': ckan_deployment_status_pods,
        'image': image,
        'latest_pod_name': latest_pod_name,
        'latest_operator_timestamp': latest_operator_timestamp,
        'helm_app_statuses': item_app_statuses,
        'helm_metadata': {
            'ckan_instance_id': instance_id,
            'namespace': instance_id,
            'status_generated_at': datetime.datetime.now(),
            'status_generated_from': subprocess.check_output(["hostname"]).decode().strip(),
        }
    }


def get_backend_url(instance_id, instance):
    return f'http://nginx.{instance_id}:8080'


def pre_update_hook(instance_id, instance, override_spec, skip_route=False, dry_run=False):
    _init_namespace(instance_id, dry_run=dry_run)
    _pre_update_hook_override_spec(override_spec, instance)
    if not instance['spec'].get('useCentralizedInfra'):
        logs.warning('Forcing centralized infra even though useCentralizedInfra is disabled')
        _pre_update_hook_modify_spec(instance_id, instance, lambda i: i.update(useCentralizedInfra=True),
                                     dry_run=dry_run)
    res = {}
    sub_domain, root_domain = _pre_update_hook_route(instance_id, skip_route, instance, res, dry_run=dry_run)
    _pre_update_hook_admin_user(instance, sub_domain, root_domain, instance_id, res, dry_run=dry_run)
    return res


def create_ckan_admin_user(instance_id, instance, user):
    pod_name = None
    while not pod_name:
        try:
            pod_name = kubectl.get_deployment_pod_name('ckan', instance_id, use_first_pod=True, required_phase='Running')
            break
        except Exception as e:
            logs.warning('Failed to find running ckan pod', str(e))
        time.sleep(20)

    name, password, email = [user[k] for k in ['name', 'password', 'email']]
    logs.info(f'Creating CKAN admin user with {name} ({email}) and {password} on pod {pod_name}')
    logs.subprocess_check_call(
        f'echo y | kubectl -n {instance_id} exec -i {pod_name} -- ckan-paster --plugin=ckan sysadmin -c /etc/ckan/production.ini add {name} password={password} email={email}',
        shell=True
    )


def _init_ckan_infra_secret(instance_id, dry_run=False):
    logs.debug('Initializing ckan infra secret', instance_id=instance_id)
    ckan_infra = config_manager.get(secret_name='ckan-infra', namespace=instance_id, required=False)
    if ckan_infra:
        logs.info('ckan-infra secret already exists')
    else:
        admin_user, admin_password, db_name = db_manager.get_admin_db_credentials()
        db_host, db_port = db_manager.get_internal_unproxied_db_host_port()
        assert int(db_port) == 5432
        logs.debug('Creating ckan-infra secret', admin_user=admin_user, admin_password=admin_password, db_name=db_name,
                   db_host=db_host, db_port=db_port)
        config_manager.set(
            values={
                'POSTGRES_HOST': db_host,
                'POSTGRES_PASSWORD': admin_password,
                'POSTGRES_USER': admin_user
            },
            secret_name='ckan-infra',
            namespace=instance_id,
            dry_run=dry_run
        )

def _create_private_container_registry_secret(instance_id):
    if config_manager.get('private-registry', secret_name='ckan-docker-registry') == 'y':
        docker_server, docker_username, docker_password, docker_email = ckan_manager.get_docker_credentials()
        image_pull_secret_name = config_manager.get('docker-image-pull-secret-name', secret_name='ckan-docker-registry')
        subprocess.call(f'kubectl -n {instance_id} create secret docker-registry {image_pull_secret_name} '
                              f'--docker-password={docker_password} '
                              f'--docker-server={docker_server} '
                              f'--docker-username={docker_username} '
                              f'--docker-email={docker_email}', shell=True)


def _init_namespace(instance_id, dry_run=False):
    logs.debug('Initializing helm-based instance deployment namespace', namespace=instance_id)
    if kubectl.get('ns', instance_id, required=False):
        logs.info(f'instance namespace already exists ({instance_id})')
    else:
        logs.info(f'creating instance namespace ({instance_id})')
        kubectl.apply(kubectl.get_resource('v1', 'Namespace', instance_id, {}), dry_run=dry_run)
        service_account_name = f'ckan-{instance_id}-operator'
        logs.debug('Creating service account', service_account_name=service_account_name)
        if not dry_run:
            kubectl_rbac_driver.update_service_account(f'ckan-{instance_id}-operator', {}, namespace=instance_id)
        role_name = f'ckan-{instance_id}-operator-role'
        logs.debug('Creating role and binding to the service account', role_name=role_name)
        if not dry_run:
            kubectl_rbac_driver.update_role(role_name, {}, [
                {
                    "apiGroups": [
                        "*"
                    ],
                    "resources": [
                        'secrets', 'pods', 'pods/exec', 'pods/portforward'
                    ],
                    "verbs": [
                        "list", "get", "create"
                    ]
                }
            ], namespace=instance_id)
            kubectl_rbac_driver.update_role_binding(
                name=f'ckan-{instance_id}-operator-rolebinding',
                role_name=f'ckan-{instance_id}-operator-role',
                namespace=instance_id,
                service_account_name=f'ckan-{instance_id}-operator',
                labels={}
            )


def _init_solr(instance_id, solr_schema, dry_run=False):
    logs.debug('Initializing solr', instance_id=instance_id)
    solr_status = solr_manager.get_collection_status(instance_id)
    logs.debug_yaml_dump(solr_status)
    if not solr_status['ready']:
        logs.info('Creating solr collection', collection_name=instance_id, solr_config=solr_schema)
        if not dry_run:
            solr_manager.create_collection(instance_id, solr_schema)
    else:
        logs.info(f'collection already exists ({instance_id})')
    solr_url = solr_status['solr_http_endpoint']
    logs.debug(solr_url=solr_url)
    assert solr_url.startswith('http') and solr_url.endswith('/solr'), f'invalid solr_url ({solr_url})'
    host, port = solr_url.replace('https://', '').replace('http://', '').replace('/solr', '').split(':')
    logs.debug('Solr initialization completed successfully', host=host, port=port)
    return host, port


def _check_instance_events(instance_id):
    status = get(instance_id)
    errors = []
    ckan_cloud_logs = []
    ckan_cloud_events = set()
    pod_names = []
    for app, app_status in status.get('helm_app_statuses', {}).items():
        for kind, kind_items in app_status.items():
            for item in kind_items:
                for error in item.get("errors", []):
                    errors.append(dict(error, kind=kind, app=app, name=item.get("name")))
                for logdata in item.get("ckan-cloud-logs", []):
                    ckan_cloud_logs.append(dict(logdata, kind=kind, app=app, name=item.get("name")))
                    if "event" in logdata:
                        ckan_cloud_events.add(logdata["event"])
                if kind == "pods":
                    pod_names.append(item["name"])
    expected_events = {
        ("ckan-env-vars-created", "ckan-env-vars-exists"),
        ("ckan-secrets-created", "ckan-secrets-exists"),
        "got-ckan-secrets",
        "ckan-entrypoint-initialized",
        "ckan-entrypoint-db-init-success",
        "ckan-entrypoint-extra-init-success"
    }
    missing = set()
    for events in expected_events:
        if isinstance(events, str):
            events = (events,)
        found = False
        for e in events:
            if e in ckan_cloud_events:
                found = True
                break
        if not found:
            missing.add('/'.join(events))
    logs.debug(ckan_cloud_events=ckan_cloud_events)
    return missing, errors, ckan_cloud_logs

def _wait_instance_events(instance_id):
    import logging
    start_time = datetime.datetime.now()
    last_message = 0
    logs.info('Waiting for instance events', start_time=start_time)
    missing_events = None
    while True:
        time.sleep(15)
        currently_missing, errors, ckan_logs = _check_instance_events(instance_id)
        if len(currently_missing) == 0:
            logs.info('All instance events completed successfully')
            break
        if currently_missing != missing_events:
            missing_events = currently_missing
            logs.info('Still waiting for', repr(sorted(missing_events)))
            start_time = datetime.datetime.now()
        time_passed = (datetime.datetime.now() - start_time).total_seconds()
        if time_passed - last_message >= 60:
            logs.info('%d seconds since started waiting' % time_passed)
            last_message += 60
        if time_passed > int(os.environ.get('CCO_WAIT_TIMEOUT', 500)):
            failed_pods = [
                item for item in kubectl.get(f'pods -n {instance_id}').get('items', [])
                    if not all(stat.get('ready') for stat in item['status']['containerStatuses'])
            ]
            logs.info('*** SOMETHING WENT WRONG!!! ***')
            logs.info(100*'#')
            logs.info(100*'#')
            if not len(failed_pods):
                logs.info('But we could not get failing containers')
                logs.info('You may try increasing default wait timeout by setting CCO_WAIT_TIMEOUT environment variable [default: 500]')
                ckan_pod_name = [
                    item['metadata']['name'] for item in kubectl.get(f'pods -n {instance_id}').get('items', [])
                        if item.get('metadata', {}).get('labels', {}).get('app') == 'ckan'
                ][0]
                _log_container_error('CONTAINER LOGS', ckan_pod_name, 'ckan')
                kubectl.call(f'logs {ckan_pod_name}', namespace=instance_id)

            logs.info('Numbe of Failed Pods: %s' % len(failed_pods))
            for pod_meta in failed_pods:
                init_containers = pod_meta['status'].get('initContainerStatuses')
                pod_name = pod_meta['metadata']['name']
                if init_containers is not None:
                    logs.info('Checking Init Containers in %s' % pod_name)
                    for i, init_container in enumerate(init_containers):
                        if not init_container.get('ready'):
                            container_name = pod_meta['spec']['initContainers'][i]['name']
                            _log_container_error('INIT CONTAINER LOGS', pod_name, container_name)
                            kubectl.call(f'logs {pod_name} -c {container_name}', namespace=instance_id)
                        else:
                            logs.info('Init Containers are fine in %s' % pod_name)
                logs.info('Checking Containers in %s' % pod_name)
                container_stats = pod_meta['status'].get('containerStatuses')
                no_log_statuses = ['PodInitializing']
                pod_status = [
                    stat.get('state', {}).get('waiting', {}).get('reason') in no_log_statuses for stat in container_stats
                ]
                if all(pod_status):
                    logs.info('Pod %s looks good describing:' % pod_name)
                    _log_container_error('KUBECTL DESCRIBE POD', pod_name)
                    kubectl.call(f'describe pod {pod_name}', namespace=instance_id)
                else:
                    logs.info('Pod %s look did not start:' % pod_name)
                    for i, container in enumerate(container_stats):
                        container_name = pod_meta['spec']['containers'][i]['name']
                        _log_container_error('CONTAINER LOGS', pod_name, container_name)
                        kubectl.call(f'logs {pod_name}', namespace=instance_id)

            logs.info(100*'#')
            logs.info(100*'#')
            raise Exception('timed out waiting for instance events')


def _pre_update_hook_admin_user(instance, sub_domain, root_domain, instance_id, res, dry_run=False):
    ckan_admin_email = instance['spec'].get('ckanAdminEmail')
    if not ckan_admin_email:
        ckan_admin_email = f'admin@{sub_domain}.{root_domain}'
    ckan_admin_password = config_manager.get(key='CKAN_ADMIN_PASSWORD', secret_name='ckan-admin-password',
                                             namespace=instance_id, required=False)
    if ckan_admin_password:
        logs.info('using existing ckan admin user')
        res['ckan-admin-password'] = ckan_admin_password
    else:
        logs.info('Will create new ckan admin user', ckan_admin_email=ckan_admin_email)
        res['ckan-admin-email'] = ckan_admin_email
        res['ckan-admin-password'] = ckan_admin_password = binascii.hexlify(os.urandom(8)).decode()
        config_manager.set(key='CKAN_ADMIN_PASSWORD', value=ckan_admin_password, secret_name='ckan-admin-password',
                           namespace=instance_id,
                           dry_run=dry_run)


def _pre_update_hook_route(instance_id, skip_route, instance, res, dry_run=False):
    root_domain = routers_manager.get_default_root_domain()
    sub_domain = instance['spec'].get('sub-domain', f'ckan-cloud-{instance_id}')
    if not skip_route:
        # full domain to route to the instance
        instance_domain = instance['spec'].get('domain')
        if instance_domain and instance_domain != f'{sub_domain}.{root_domain}':
            logs.warning(f'instance domain was changed from {instance_domain} to {sub_domain}.{root_domain}')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(domain=f'{sub_domain}.{root_domain}'),
                                         dry_run=dry_run)
        # instance is added to router only if this is true, as all routers must use SSL and may use sans SSL too
        with_sans_ssl = instance['spec'].get('withSansSSL')
        if not with_sans_ssl:
            logs.warning(f'forcing with_sans_ssl, even though withSansSSL is disabled')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(withSansSSL=True),
                                         dry_run=dry_run)
        # subdomain to register on the default root domain
        register_subdomain = instance['spec'].get('registerSubdomain')
        if register_subdomain != sub_domain:
            logs.warning(f'instance register sub domain was changed from {register_subdomain} to {sub_domain}')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(registerSubdomain=sub_domain),
                                         dry_run=dry_run)
        res.update(**{'root-domain': root_domain, 'sub-domain': sub_domain})
        site_url = instance['spec'].get('siteUrl')
        if site_url != f'https://{sub_domain}.{root_domain}':
            logs.warning(f'instance siteUrl was changed from {site_url} to https://{sub_domain}.{root_domain}')
            _pre_update_hook_modify_spec(instance_id, instance,
                                         lambda i: i.update(siteUrl=f'https://{sub_domain}.{root_domain}'),
                                         dry_run=dry_run)
    return sub_domain, root_domain


def _pre_update_hook_override_spec(override_spec, instance):
    # applies override spec, but doesn't persist
    if override_spec:
        for k, v in override_spec.items():
            logs.info(f'Applying override spec {k}={v}')
            instance['spec'][k] = v


def _pre_update_hook_modify_spec(instance_id, instance, callback, dry_run=False):
    # applies changes to both the non-persistent spec and persists the changes on latest instance spec
    latest_instance = crds_manager.get(INSTANCE_CRD_SINGULAR, crds_manager.get_resource_name(
        INSTANCE_CRD_SINGULAR, instance_id
    ), required=True)
    callback(instance['spec'])
    callback(latest_instance['spec'])
    kubectl.apply(latest_instance, dry_run=dry_run)

def _scale_down_scale_up(deployment='ckan', namespace=None, replicas=1):
    logs.info('Scaling ckan replicas')
    kubectl.call(f'scale deployment {deployment} --replicas=0', namespace=namespace)
    kubectl.call(f'scale deployment {deployment} --replicas={replicas}', namespace=namespace)
    time.sleep(20)

def _log_container_error(text, pod_name, container_name=None):
    heading = f"****** {text} ******"
    logs.info()
    logs.info(heading)
    logs.info(f'Pod: {pod_name} - Container: {container_name}')
    logs.info("-" * len(heading))
    logs.info()



from ckan_cloud_operator import kubectl

from ckan_cloud_operator.providers.ckan.instance import manager as ckan_instance_manager


from ckan_cloud_operator.routers.routes.backend_url_subdomain import (
    get_frontend_hostname,
    get_domain_parts,
    get_default_root_domain,
    get_route,
    _init_route,
    pre_deployment_hook
)


def get_backend_url(route):
    name, spec = _init_route(route)
    return ckan_instance_manager.get_backend_url(spec['ckan-instance-id'])



import subprocess
import yaml
import datetime
import time
from ckan_cloud_operator import kubectl


class DeisCkanInstanceDeployment(object):

    def __init__(self, instance):
        self.instance = instance

    def update(self):
        if not self.instance.annotations.update_status('deployment', 'created', lambda: self._deploy()):
            self._deploy()

    def delete(self):
        print(f'Deleting instance {self.instance.id}')
        subprocess.check_call(f'kubectl -n {self.instance.id} delete deployment/{self.instance.id} --now', shell=True)

    def get(self):
        exitcode, output = subprocess.getstatusoutput(f'kubectl -n {self.instance.id} get deployment/{self.instance.id} -o yaml')
        if exitcode == 0:
            deployment = yaml.load(output)
            status = kubectl.get_item_detailed_status(deployment)
            ready = len(status.get('error', [])) == 0
            status['pods'] = []
            pods = kubectl.get('pods -l app=ckan', namespace=self.instance.id, required=False)
            image = None
            latest_operator_timestamp, latest_pod_name, latest_pod_status = None, None, None
            if pods:
                for pod in pods['items']:
                    pod_operator_timestamp = pod['metadata']['annotations']['ckan-cloud/operator-timestamp']
                    if not latest_operator_timestamp or latest_operator_timestamp < pod_operator_timestamp:
                        latest_operator_timestamp = pod_operator_timestamp
                        latest_pod_name = pod['metadata']['name']
                    pod_status = kubectl.get_item_detailed_status(pod)
                    status_code, output = subprocess.getstatusoutput(
                        f'kubectl -n {self.instance.id} logs {pod["metadata"]["name"]} -c ckan --tail 5',
                    )
                    if status_code == 0:
                        pod_status['logs'] = output
                    else:
                        pod_status['logs'] = None
                    if not image:
                        image = pod["spec"]["containers"][0]["image"]
                    else:
                        if image != pod["spec"]["containers"][0]["image"]:
                            ready = False
                            image = pod["spec"]["containers"][0]["image"]
                    status['pods'].append(pod_status)
                    if latest_pod_name == pod_status['name']:
                        latest_pod_status = pod_status
                if not latest_pod_status or len(latest_pod_status.get('errors', [])) > 0 or latest_pod_status['logs'] is None:
                    ready = False
            else:
                ready = False
            return dict(status, ready=ready, image=image, latest_pod_name=latest_pod_name, latest_operator_timestamp=latest_operator_timestamp)
        else:
            return {'ready': False, 'error': output}

    def _deploy(self):
        print(f'Deploying instance {self.instance.id}')
        ckanContainerSpec = dict(
            name='ckan',
            envFrom=[{'secretRef': {'name': 'ckan-envvars'}}],
            readinessProbe={
                'failureThreshold': 1,
                'initialDelaySeconds': 5,
                'periodSeconds': 5,
                'successThreshold': 1,
                'tcpSocket': {
                    'port': 5000
                },
                'timeoutSeconds': 5
            },
            resources={
                'limits': {
                    'memory': '2Gi',
                },
                'requests': {
                    'memory': '1Gi',
                }
            }
        )
        ckanContainerSpec.update(self.instance.spec.spec['ckanContainerSpec'])
        if 'imageFromGitlab' in ckanContainerSpec:
            ckanContainerSpec['image'] = 'registry.gitlab.com/{}:latest'.format(ckanContainerSpec.pop('imageFromGitlab'))
        imagePullSecrets = ckanContainerSpec.pop('imagePullSecrets', None)
        ckanPodSpec = dict(self.instance.spec.spec['ckanPodSpec'],
                           serviceAccountName=f'ckan-{self.instance.id}-operator',
                           containers=[ckanContainerSpec])
        if imagePullSecrets:
            ckanPodSpec['imagePullSecrets'] = imagePullSecrets
        else:
            ckanPodSpec['imagePullSecrets'] = [{'name': f'{self.instance.id}-registry'}]
        deployment = {'apiVersion': 'apps/v1beta1',
                      'kind': 'Deployment',
                      'metadata': {
                          'name': self.instance.id,
                          'namespace': self.instance.id
                      },
                      'spec': {
                          'replicas': 1,
                          'revisionHistoryLimit': 10,
                          'progressDeadlineSeconds': 600,
                          'strategy': {
                              'type': 'RollingUpdate',
                              'rollingUpdate': {
                                  'maxSurge': 1,
                                  'maxUnavailable': 0
                              }
                          },
                          'template': {
                              'metadata': {
                                  'labels': {
                                      'app': 'ckan'
                                  },
                                  'annotations': {
                                      'ckan-cloud/operator-timestamp': str(datetime.datetime.now())
                                  }
                              },
                              'spec': ckanPodSpec
                          }
                      }}
        subprocess.run('kubectl apply -f -', input=yaml.dump(deployment).encode(), shell=True, check=True)



import os
from dataflows import Flow, load, printer, checkpoint, dump_to_path, add_field
from ckan_cloud_operator.gitlab import CkanGitlab


def get_gitlab_repo(rows):
    if rows.res.name == 'ckan-cloud-instances':
        for row in rows:
            image = row['image']
            if image.endswith(' (imageFromGitlab)'):
                gitlab_repo = image.replace(' (imageFromGitlab)', '')
            elif image.startswith('registry.gitlab.com/viderum/cloud-'):
                gitlab_repo = image.replace('registry.gitlab.com/', '').split('@')[0]
            else:
                gitlab_repo = None
            yield dict(row, gitlab_repo=gitlab_repo)
    else:
        yield from rows


def parse_dockerfiles():
    gitlab_repos = {}

    def _parse_gitlab_repos(rows):
        if rows.res.name == 'ckan-cloud-instances':
            for row in rows:
                gitlab_repo = row['gitlab_repo']
                if gitlab_repo in gitlab_repos:
                    gitlab_repos[gitlab_repo]['instances'].append(row)
                else:
                    gitlab_repos[gitlab_repo] = {'instances': [row]}
                yield row
        else:
            yield from rows

    def _get_dockerfile_from(dockerfile):
        if dockerfile:
            return [line.replace('FROM ', '') for line in dockerfile.split('\n') if line.startswith('FROM')][0]
        else:
            return None

    def _parse_ckan_extensions(rows):
        if rows.res.name == 'dockerfiles':
            for row in rows:
                row['ckan_exts'] = []
                if row['dockerfile']:
                    for line in row['dockerfile'].split('\n'):
                        if 'https://github.com/' in line and '.git@' in line and '#egg=' in line:
                            ext = line.split('https://github.com/')[1].split('#egg=')[0].replace('.git@', '@')
                            row['ckan_exts'].append(ext)
                            if 'ckanext-s3filestore' in ext:
                                row['ckanext-s3filestore'] = ext
                yield row
        else:
            yield from rows

    def _get_dockerfile_row(gitlab_repo_name, gitlab_repo):
        try:
            dockerfile = CkanGitlab()._get_file(gitlab_repo_name, 'Dockerfile')
        except Exception:
            dockerfile = None
        return {
            'gitlab_repo': gitlab_repo_name,
            'instances': [i['name'] for i in gitlab_repo['instances']],
            'from': _get_dockerfile_from(dockerfile),
            'dockerfile': dockerfile
        }

    def _parse_dockerfiles(package):
        package.pkg.add_resource({'name': 'dockerfiles', 'path': 'dockerfiles.csv', 'schema': {'fields': [
            {'name': 'gitlab_repo', 'type': 'string'},
            {'name': 'instances', 'type': 'array'},
            {'name': 'from', 'type': 'string'},
            {'name': 'dockerfile', 'type': 'string'}
        ]}})
        yield package.pkg
        yield from package
        yield (_get_dockerfile_row(gitlab_repo_name, gitlab_repo) for gitlab_repo_name, gitlab_repo in
               gitlab_repos.items())

    return Flow(
        _parse_gitlab_repos,
        _parse_dockerfiles,
        checkpoint('ckan_images_dockerfiles'),
        add_field('ckan_exts', 'array'),
        add_field('ckanext-s3filestore', 'string'),
        _parse_ckan_extensions,
    )


def main_flow(prefix, operator):
    return Flow(
        load(f'data/{prefix}/resources/datapackage.json', resources=['ckan-cloud-instances']),
        add_field('gitlab_repo', 'string'),
        get_gitlab_repo,
        parse_dockerfiles(),
    )


if __name__ == '__main__':
    prefix = os.environ['DATAPACKAGE_PREFIX']
    operator = os.environ.get('CKAN_CLOUD_OPERATOR_BIN', 'ckan-cloud-operator')
    Flow(
        main_flow(prefix, operator),
        printer(num_rows=1, fields=['name', 'image', 'gitlab_repo', 'from']),
        dump_to_path(f'data/{prefix}/ckan_images')
    ).process()



import json
import unittest
from unittest.mock import patch, call

from ckan_cloud_operator.providers.storage.gcloud import manager as gcloud_manager
from ckan_cloud_operator.providers.storage.s3 import manager as s3_manager


class S3ManagerTestCase(unittest.TestCase):
    @patch('ckan_cloud_operator.providers.storage.s3.manager.get_aws_credentials')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    @patch('ckan_cloud_operator.providers.storage.s3.manager._generate_password')
    def test_create_bucket(self, _generate_password, aws_check_output, list_s3_buckets, get_aws_credentials):
        aws_check_output.return_value = ''
        list_s3_buckets.return_value = []
        get_aws_credentials.return_value = {'region': 'us-west-1'}
        _generate_password.return_value = 'abcdefabcdefabcdef'

        result = s3_manager.create_bucket('new-instance')

        get_aws_credentials.assert_called_once()
        list_s3_buckets.assert_called_once()
        aws_check_output.assert_called_once_with('s3 mb s3://new-instance-ccabcdefabcdefabcdef --region us-west-1')
        expected_result = {
            'BUCKET_NAME': 's3://new-instance-ccabcdefabcdefabcdef',
            'BUCKET_ACCESS_KEY': None,
            'BUCKET_ACCESS_SECRET': None
        }
        self.assertEqual(result, expected_result)

    @patch('ckan_cloud_operator.providers.storage.s3.manager.get_aws_credentials')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    @patch('ckan_cloud_operator.providers.storage.s3.manager._generate_password')
    def test_create_bucket_that_already_exists(self, _generate_password, aws_check_output, list_s3_buckets, get_aws_credentials):
        aws_check_output.return_value = ''
        list_s3_buckets.return_value = ['new-instance']
        get_aws_credentials.return_value = {'region': 'us-west-1'}
        _generate_password.return_value = 'abcdefabcdefabcdef'

        with self.assertRaisesRegex(Exception, 'Bucket for this instance already exists'):
            s3_manager.create_bucket('new-instance')

        expected_result = {
            'BUCKET_NAME': 's3://new-instance-ccabcdefabcdefabcdef',
            'BUCKET_ACCESS_KEY': None,
            'BUCKET_ACCESS_SECRET': None
        }
        self.assertEqual(s3_manager.create_bucket('new-instance', exists_ok=True), expected_result)
        aws_check_output.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.s3.manager.get_aws_credentials')
    def test_create_bucket_without_region(self, get_aws_credentials):
        get_aws_credentials.return_value = {}
        with self.assertRaisesRegex(AssertionError, 'No default region set for the cluster'):
            s3_manager.create_bucket('new-instance')

    @patch('ckan_cloud_operator.providers.storage.s3.manager.get_aws_credentials')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    @patch('ckan_cloud_operator.providers.storage.s3.manager._generate_password')
    def test_create_bucket_dry_run(self, _generate_password, aws_check_output, list_s3_buckets, get_aws_credentials):
        aws_check_output.return_value = ''
        list_s3_buckets.return_value = []
        get_aws_credentials.return_value = {'region': 'us-west-1'}
        _generate_password.return_value = 'abcdefabcdefabcdef'

        result = s3_manager.create_bucket('new-instance', dry_run=True)

        get_aws_credentials.assert_called_once()
        list_s3_buckets.assert_called_once()
        aws_check_output.assert_not_called()

        expected_result = {
            'BUCKET_NAME': 's3://new-instance-ccabcdefabcdefabcdef',
            'BUCKET_ACCESS_KEY': None,
            'BUCKET_ACCESS_SECRET': None
        }
        self.assertEqual(result, expected_result)

    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.kubectl.get')
    def test_delete_bucket(self, kubectl_get, aws_check_output, list_s3_buckets):
        list_s3_buckets.return_value = ['new-instance-ccabcdefabcdefabcdef']
        kubectl_get.return_value = {
            'spec': {
                'ckanStorageBucket': {
                    's3': {
                        'BUCKET_NAME': 's3://new-instance-ccabcdefabcdefabcdef'
                    }
                }
            }
        }

        s3_manager.delete_bucket('new-instance')

        expected_calls = [
            call('s3 rm s3://new-instance-ccabcdefabcdefabcdef --recursive'),
            call('s3 rb s3://new-instance-ccabcdefabcdefabcdef')
        ]
        self.assertEqual(aws_check_output.call_count, 2)
        self.assertEqual(aws_check_output.mock_calls, expected_calls)

    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    def test_delete_bucket_that_does_not_exist(self, aws_check_output, list_s3_buckets):
        list_s3_buckets.return_value = []

        s3_manager.delete_bucket('new-instance')

        aws_check_output.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.kubectl.get')
    def test_delete_bucket_dry_run(self, kubectl_get, aws_check_output, list_s3_buckets):
        list_s3_buckets.return_value = ['new-instance-ccabcdefabcdefabcdef']
        kubectl_get.return_value = {
            'spec': {
                'ckanStorageBucket': {
                    's3': {
                        'BUCKET_NAME': 's3://new-instance-ccabcdefabcdefabcdef'
                    }
                }
            }
        }

        s3_manager.delete_bucket('new-instance', dry_run=True)

        aws_check_output.assert_called_once_with('s3 rm s3://new-instance-ccabcdefabcdefabcdef --recursive --dryrun')

    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.kubectl.get')
    def test_get_bucket(self, kubectl_get, list_s3_buckets):
        list_s3_buckets.return_value = ['new-instance-ccabcdefabcdefabcdef']
        kubectl_get.return_value = {
            'spec': {
                'ckanStorageBucket': {
                    's3': 's3://new-instance-ccabcdefabcdefabcdef'
                }
            }
        }

        expected_result = {
            'instance_id': 'new-instance',
            'ckanStorageBucket': 's3://new-instance-ccabcdefabcdefabcdef'
        }
        self.assertEqual(s3_manager.get_bucket('new-instance'), expected_result)

    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.kubectl.get')
    def test_get_bucket_that_does_not_present_in_s3(self, kubectl_get, list_s3_buckets):
        list_s3_buckets.return_value = ['old-instance']

        self.assertIsNone(s3_manager.get_bucket('new-instance'))
        kubectl_get.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.s3.manager.list_s3_buckets')
    @patch('ckan_cloud_operator.providers.storage.s3.manager.kubectl.get')
    def test_get_bucket_when_instance_has_no_s3_buckets(self, kubectl_get, list_s3_buckets):
        list_s3_buckets.return_value = ['new-instance']
        kubectl_get.return_value = {
            'spec': {
               'ckanStorageBucket': {
                    'gcloud': 'new-instance'
                }
            }
        }

        self.assertIsNone(s3_manager.get_bucket('new-instance'))

    @patch('ckan_cloud_operator.providers.storage.s3.manager.kubectl.get')
    def test_list_buckets(self, kubectl_get):
        kubectl_get.return_value = {
            'items': [
                {
                    'spec': {
                        'id': 'new-instance',
                        'ckanStorageBucket': {
                            's3': 's3://new-instance'
                        }
                    }
                },
                {
                    'spec': {
                        'id': 'old-instance',
                        'ckanStorageBucket': {
                            's3': 's3://old-instance'
                        }
                    }
                },
                {
                    'spec': {
                        'id': 'gcloud-instance',
                        'ckanStorageBucket': {
                            'gcloud': 'gcloud-instance'
                        }
                    }
                },
                {
                    'spec': {
                        'no-data': 'invalid spec'
                    }
                }
            ]
        }

        expected_result = [
            {
                'instance_id': 'new-instance',
                'ckanStorageBucket': 's3://new-instance'
            },
            {
                'instance_id': 'old-instance',
                'ckanStorageBucket': 's3://old-instance'
            }
        ]
        self.assertEqual(s3_manager.list_buckets(), expected_result)
        kubectl_get.assert_called_once_with('ckancloudckaninstance')

    @patch('ckan_cloud_operator.providers.storage.s3.manager.aws_check_output')
    def test_list_s3_buckets(self, aws_check_output):
        aws_check_output.return_value = b"""
2019-11-10 21:37:53 new-instance
2019-11-10 21:32:57 old-instance
2019-11-10 21:32:57 another-instance"""

        expected_result = [
            ['2019-11-10 21:37:53', 'new-instance'],
            ['2019-11-10 21:32:57', 'old-instance'],
            ['2019-11-10 21:32:57', 'another-instance']
        ]
        self.assertEqual(s3_manager.list_s3_buckets(), expected_result)

        expected_result = ['new-instance', 'old-instance', 'another-instance']
        self.assertEqual(s3_manager.list_s3_buckets(names_only=True), expected_result)


class GCloudManagerTestCase(unittest.TestCase):
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.cluster_config_get')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_create_bucket(self, gcloud_check_output, list_s3_buckets, cluster_config_get):
        gcloud_check_output.return_value = ''
        list_s3_buckets.return_value = []
        cluster_config_get.return_value = 'europe-west2-c'

        result = gcloud_manager.create_bucket('new-instance')
        expected_result = {
            'BUCKET_NAME': 'gs://new-instance'
        }

        cluster_config_get.assert_called_once()
        list_s3_buckets.assert_called_once()
        gcloud_check_output.assert_called_once_with('mb gs://new-instance -l europe-west2-c', gsutil=True)
        self.assertEqual(result, expected_result)

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.cluster_config_get')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_create_bucket_that_already_exists(self, gcloud_check_output, list_gcloud_buckets, cluster_config_get):
        gcloud_check_output.return_value = ''
        list_gcloud_buckets.return_value = ['new-instance']
        cluster_config_get.return_value = 'europe-west2-c'

        with self.assertRaisesRegex(Exception, 'Bucket for this instance already exists'):
            gcloud_manager.create_bucket('new-instance')

        expected_result = {
            'BUCKET_NAME': 'gs://new-instance'
        }
        self.assertEqual(gcloud_manager.create_bucket('new-instance', exists_ok=True), expected_result)
        gcloud_check_output.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.cluster_config_get')
    def test_create_bucket_without_region(self, cluster_config_get):
        cluster_config_get.return_value = {}
        with self.assertRaisesRegex(AssertionError, 'No default region set for the cluster'):
            gcloud_manager.create_bucket('new-instance')

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.cluster_config_get')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_create_bucket_dry_run(self, gcloud_check_output, list_gcloud_buckets, cluster_config_get):
        gcloud_check_output.return_value = ''
        list_gcloud_buckets.return_value = []
        cluster_config_get.return_value = 'europe-west2-c'

        expected_result = {
            'BUCKET_NAME': 'gs://new-instance'
        }
        result = gcloud_manager.create_bucket('new-instance', dry_run=True)

        cluster_config_get.assert_called_once()
        list_gcloud_buckets.assert_called_once()
        gcloud_check_output.assert_not_called()
        self.assertEqual(result, expected_result)

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_delete_bucket(self, gcloud_check_output, list_gcloud_buckets):
        list_gcloud_buckets.return_value = ['new-instance']

        gcloud_manager.delete_bucket('new-instance')

        expected_calls = [
            call('rm -r gs://new-instance', gsutil=True),
        ]
        self.assertEqual(gcloud_check_output.call_count, 1)
        self.assertEqual(gcloud_check_output.mock_calls, expected_calls)

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_delete_bucket_that_does_not_exist(self, gcloud_check_output, list_gcloud_buckets):
        list_gcloud_buckets.return_value = []

        gcloud_manager.delete_bucket('new-instance')

        gcloud_check_output.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_delete_bucket_dry_run(self, gcloud_check_output, list_gcloud_buckets):
        list_gcloud_buckets.return_value = ['new-instance']

        gcloud_manager.delete_bucket('new-instance', dry_run=True)

        gcloud_check_output.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.kubectl.get')
    def test_get_bucket(self, kubectl_get, list_gcloud_buckets):
        list_gcloud_buckets.return_value = ['new-instance']
        kubectl_get.return_value = {
            'spec': {
                'ckanStorageBucket': {
                    'gcloud': 'gs://new-instance'
                }
            }
        }

        expected_result = {
            'instance_id': 'new-instance',
            'ckanStorageBucket': 'gs://new-instance'
        }
        self.assertEqual(gcloud_manager.get_bucket('new-instance'), expected_result)

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.kubectl.get')
    def test_get_bucket_that_is_not_present_in_gcp(self, kubectl_get, list_gcloud_buckets):
        list_gcloud_buckets.return_value = ['old-instance']

        self.assertIsNone(gcloud_manager.get_bucket('new-instance'))
        kubectl_get.assert_not_called()

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.list_gcloud_buckets')
    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.kubectl.get')
    def test_get_bucket_when_instance_has_no_gcp_buckets(self, kubectl_get, list_gcloud_buckets):
        list_gcloud_buckets.return_value = ['new-instance']
        kubectl_get.return_value = {
            'spec': {
                'ckanStorageBucket': {
                    's3': 'new-instance'
                }
            }
        }

        self.assertIsNone(gcloud_manager.get_bucket('new-instance'))

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.kubectl.get')
    def test_list_buckets(self, kubectl_get):
        kubectl_get.return_value = {
            'items': [
                {
                    'spec': {
                        'id': 'new-instance',
                        'ckanStorageBucket': {
                            'gcloud': 'gs://new-instance'
                        }
                    }
                },
                {
                    'spec': {
                        'id': 'old-instance',
                        'ckanStorageBucket': {
                            'gcloud': 'gs://old-instance'
                        }
                    }
                },
                {
                    'spec': {
                        'id': 's3-instance',
                        'ckanStorageBucket': {
                            's3': 's3-bucket'
                        }
                    }
                },
                {
                    'spec': {
                        'no-data': 'invalid spec'
                    }
                }
            ]
        }

        expected_result = [
            {
                'instance_id': 'new-instance',
                'ckanStorageBucket': 'gs://new-instance'
            },
            {
                'instance_id': 'old-instance',
                'ckanStorageBucket': 'gs://old-instance'
            }
        ]
        self.assertEqual(gcloud_manager.list_buckets(), expected_result)
        kubectl_get.assert_called_once_with('ckancloudckaninstance')

    @patch('ckan_cloud_operator.providers.storage.gcloud.manager.gcloud_check_output')
    def test_list_gcloud_buckets(self, gcloud_check_output):
        gcloud_check_output.return_value = b"""
gs://new-instance/
gs://old-instance/
gs://another-instance/"""

        expected_result = [
            'new-instance',
            'old-instance',
            'another-instance'
        ]
        self.assertEqual(gcloud_manager.list_gcloud_buckets(), expected_result)



import yaml
import subprocess
import binascii
import os
import sys

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs
from ckan_cloud_operator.labels import manager as labels_manager

from .constants import OPERATOR_NAMESPACE, OPERATOR_CONFIGMAP


def get_operator_version(verify=False):
    installed_image_tag = _get_installed_operator_image_tag()
    if verify:
        expected_image_tag = _get_expected_operator_image_tag()
        if not expected_image_tag:
            logs.info('No configmap created yet')
            return 'not-configured'
        if installed_image_tag and len(installed_image_tag) >= 2:
            assert installed_image_tag == expected_image_tag, \
                f'installed tag mismatch (expected={expected_image_tag}, actual={installed_image_tag})'
        else:
            logs.error("No version tag could be found. for local development, "
                       "make sure you use correct version and run the following "
                       "to create the file with correct version:\n\n"
                       f"echo {expected_image_tag} | sudo tee /etc/CKAN_CLOUD_OPERATOR_IMAGE_TAG\n")
            logs.exit_catastrophic_failure()
    return installed_image_tag


def print_info(debug=False, minimal=False):
    print(yaml.dump([dict(
        get_kubeconfig_info(),
        nodes=get_node_names(),
    )], default_flow_style=False))
    if not minimal:
        from ckan_cloud_operator.providers import manager as providers_manager
        print(yaml.dump([providers_manager.get_provider('cluster').get_info(debug=debug)], default_flow_style=False))
        if debug:
            assert all([
                subprocess.call(f'kubectl cluster-info', shell=True) == 0,
                subprocess.call(f'kubectl config get-contexts $(kubectl config current-context)', shell=True) == 0,
                subprocess.call(f'kubectl get nodes', shell=True) == 0
            ]), 'failed to get all debug info'
        else:
            return True


def initialize(log_kwargs=None, interactive=False, default_cluster_provider=None, skip_to=None):
    if interactive:
        logs.info('Starting interactive initialization of the operator on the following cluster:')
        print_info(minimal=True)
        if sys.stdout.isatty():
            input('Verify you are connected to the right cluster and press <RETURN> to continue')

    if not skip_to:
        logs.info(f'Creating operator namespace: {OPERATOR_NAMESPACE}', **(log_kwargs or {}))
        subprocess.call(f'kubectl create ns {OPERATOR_NAMESPACE}', shell=True)
        assert default_cluster_provider in ['gcloud', 'aws', 'azure', 'minikube'], f'invalid cluster provider: {default_cluster_provider}'
        subprocess.call(f'kubectl -n {OPERATOR_NAMESPACE} create secret generic ckan-cloud-provider-cluster-{default_cluster_provider}', shell=True)
        subprocess.call(f'kubectl -n {OPERATOR_NAMESPACE} create configmap operator-conf --from-literal=ckan-cloud-operator-image=viderum/ckan-cloud-operator:latest --from-literal=label-prefix={OPERATOR_NAMESPACE}', shell=True)

    from ckan_cloud_operator.providers import manager as providers_manager
    from ckan_cloud_operator.labels import manager as labels_manager
    from ckan_cloud_operator.crds import manager as crds_manager
    from ckan_cloud_operator.providers.db import manager as db_manager
    from ckan_cloud_operator.providers.ckan import manager as ckan_manager
    from ckan_cloud_operator.providers.routers import manager as routers_manager
    from ckan_cloud_operator.providers.solr import manager as solr_manager
    from ckan_cloud_operator.providers.storage import manager as storage_manager
    from ckan_cloud_operator.providers.apps import manager as apps_manager

    for component, func in (
            ('labels', lambda lk: labels_manager.initialize(log_kwargs=lk)),
            ('cluster', lambda lk: providers_manager.get_provider('cluster', default=default_cluster_provider).initialize(interactive=interactive)),
            ('crds', lambda lk: crds_manager.initialize(log_kwargs=lk)),
            ('db', lambda lk: db_manager.initialize(log_kwargs=lk, interactive=interactive, default_cluster_provider=default_cluster_provider)),
            ('routers', lambda lk: routers_manager.initialize(interactive=interactive)),
            ('solr', lambda lk: solr_manager.initialize(interactive=interactive)),
            ('storage', lambda lk: storage_manager.initialize(interactive=interactive)),
            ('ckan', lambda lk: ckan_manager.initialize(interactive=interactive)),
            ('apps', lambda lk: apps_manager.initialize(interactive=interactive)),
    ):
        if not skip_to or skip_to == component:
            skip_to = None
            log_kwargs = {'cluster-init': component}
            logs.info(f'Initializing', **log_kwargs)
            func(log_kwargs)


def get_kubeconfig_info():
    kubeconfig = kubectl.get('config view', get_cmd='')
    clusters = kubeconfig['clusters']
    num_clusters = len(clusters)
    assert num_clusters == 1, f'Invalid number of clusters in kubeconfig: {num_clusters}'
    cluster = clusters[0]
    contexts = kubeconfig['contexts']
    num_contexts = len(contexts)
    assert num_contexts == 1, f'Invalid number of contexts in kubeconfig: {num_contexts}'
    context = contexts[0]
    return {
        'name': cluster['name'],
        'server': cluster['cluster']['server'],
        'user': context['context']['user'],
        **get_kube_version_info(),
    }


def get_kube_version_info():
    version = kubectl.get('', get_cmd='version')
    version = {
        'clientMajor': version['clientVersion']['major'],
        'clientMinor': version['clientVersion']['minor'],
        'serverMajor': version['serverVersion']['major'],
        'serverMinor': version['serverVersion']['minor'],
    }
    try:
        client_major = int(version['clientMajor'])
    except Exception:
        raise Exception(f'Failed to get kubectl client major version (clientMajor={version["clientMajor"]}')
    client_minor = version['clientMinor']
    if client_minor.endswith('+'):
        client_minor = int(client_minor[:-1])
    else:
        client_minor = int(client_minor)
    assert client_major == 1 and client_minor >= 11, 'Invalid kubectl client version, ' \
                                                     'minimal supported version: 1.11\n' \
                                                     'If you are using GKE, run: gcloud components update'
    try:
        server_major = int(version['serverMajor'])
    except Exception:
        raise Exception(f'Failed to get Kubernetes server major version (serverMajor={version["serverMajor"]}')
    server_minor = version['serverMinor']
    if server_minor.endswith('+'):
        server_minor = int(server_minor[:-1])
    else:
        server_minor = int(server_minor)
    assert server_major == 1 and server_minor >= 10, "Invalid Kubernetes server version, " \
                                                     "minimal supported version: 1.10"
    return version


def get_node_names():
    return kubectl.check_output('get nodes -o custom-columns=name:.metadata.name --no-headers').decode().splitlines()


def get_operator_namespace_name():
    return OPERATOR_NAMESPACE


def get_operator_configmap_name():
    return OPERATOR_CONFIGMAP


def get_operator_configmap_namespace_defaults(configmap_name=None, namespace=None):
    return (configmap_name or get_operator_configmap_name()), (namespace or get_operator_namespace_name())


def get_cluster_name():
    from ckan_cloud_operator.providers import manager as providers_manager
    return providers_manager.get_provider('cluster').get_name()


def get_cluster_kubeconfig_spec():
    from ckan_cloud_operator.providers import manager as providers_manager
    return providers_manager.get_provider('cluster').get_cluster_kubeconfig_spec()


def get_provider():
    from ckan_cloud_operator.providers import manager as providers_manager
    return providers_manager.get_provider('cluster')


def get_provider_id():
    from ckan_cloud_operator.providers import manager as providers_manager
    return providers_manager.get_provider_id('cluster', default='gcloud')


def create_volume(disk_size_gb, labels, use_existing_disk_name=None, zone=0):
    assert len(labels) > 0, 'must provide some labels to identify the volume'
    labels = dict(
        labels,
        **labels_manager.get_resource_labels(label_suffixes=_get_cluster_volume_label_suffixes())
    )
    return get_provider().create_volume(disk_size_gb, labels, use_existing_disk_name=use_existing_disk_name, zone=zone)


def get_or_create_multi_user_volume_claim(label_suffixes):
    assert len(label_suffixes) > 0, 'must provide some labels to identify the volume'
    claim_labels = labels_manager.get_resource_labels(label_suffixes=dict(
        label_suffixes, **_get_cluster_volume_label_suffixes()
    ))
    pvcs = kubectl.get_items_by_labels('PersistentVolumeClaim', claim_labels, required=False)
    if len(pvcs) > 0:
        assert len(pvcs) == 1
        claim_name = pvcs[0]['metadata']['name']
    else:
        storage_class_name = get_multi_user_storage_class_name()
        claim_name = 'cc' + _generate_password(12)
        logs.info(f'Creating persistent volume claim: {claim_name}')
        logs.info(f'labels: {claim_labels}')
        kubectl.apply(kubectl.get_persistent_volume_claim(
            claim_name,
            claim_labels,
            {
                'storageClassName': storage_class_name,
                'accessModes': ['ReadWriteOnce' if get_provider_id() == 'azure' else 'ReadWriteMany'],
                'resources': {
                    'requests': {
                        'storage': '1Mi'
                    }
                }
            }
        ))
    return {'persistentVolumeClaim': {'claimName': claim_name}}


def get_multi_user_storage_class_name():
    return 'cca-ckan'


def provider_exec(cmd):
    get_provider().exec(cmd)


def setup_autoscaler(expander='random', min_nodes=1, max_nodes=10, zone='', node_pool='default-pool'):
    from ckan_cloud_operator.drivers.helm import driver as helm_driver
    from ckan_cloud_operator.providers import manager as providers_manager

    cloud_provider = get_provider_id()
    cluster_name = kubectl.check_output('config current-context').decode().replace('\n', '')

    if cloud_provider == 'gcloud':
        """GKE has built-in autoscaler"""
        from ckan_cloud_operator import gcloud

        gcloud.check_call(f'container clusters update {cluster_name} --enable-autoscaling --min-nodes {min_nodes} --max-nodes {max_nodes} --zone {zone} --node_pool {node_pool}')
        return

    values = {
        'image.tag': 'v1.13.1',
        'autoDiscovery.clusterName': cluster_name,
        'extraArgs.balance-similar-node-groups': 'false',
        'extraArgs.expander': expander,
        'cloudProvider': cloud_provider,
        'rbac.create': 'true'
    }
    if values['cloudProvider'] == 'aws':
        zone = providers_manager.get_provider('cluster').get_info().get('zone')
        while not zone:
            zone = input('Enter the AWS cluster region: ')
        values['awsRegion'] = zone.strip()

    helm_driver.deploy(
        tiller_namespace='kube-system',
        chart_repo='https://kubernetes-charts.storage.googleapis.com',
        chart_name='stable/cluster-autoscaler',
        chart_version='',
        release_name='cluster-autoscaler',
        namespace='kube-system',
        chart_repo_name='stable',
        values=values
    )

def _generate_password(l):
    return binascii.hexlify(os.urandom(l)).decode()


def _get_cluster_volume_label_suffixes():
    return {'provider-cluster-volume': 'multi-user'}


def _get_installed_operator_image_tag():
    if os.path.exists('/etc/CKAN_CLOUD_OPERATOR_IMAGE_TAG'):
        with open('/etc/CKAN_CLOUD_OPERATOR_IMAGE_TAG') as f:
            return f.read().strip()
    else:
        return None


def _get_expected_operator_image_tag():
    from ckan_cloud_operator.config import manager as config_manager
    expected_image = config_manager.get('ckan-cloud-operator-image')
    if not expected_image:
        return
    assert '@' not in expected_image and ':' in expected_image, f'invalid expected image: {expected_image}'
    _, expected_image_tag = expected_image.split(':')
    return expected_image_tag



import click

from ckan_cloud_operator import logs

from . import manager


@click.group()
def helm():
    """Manage CKAN instances deployed via Helm"""
    pass


@helm.command()
def initialize():
    manager.initialize()



import json

from ckan_cloud_operator import logs, kubectl
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.providers.cluster.azure.manager import _config_get as cluster_config_get, az_check_output

from ..constants import CONFIG_NAME
from .constants import PROVIDER_ID


def initialize(interactive=False, storage_suffix='', use_existing_disk_name=False, dry_run=False):
    default_zone = cluster_config_get('azure-default-location')
    assert default_zone, 'No cluster region specified.'
    if interactive and not dry_run:
        config_manager.interactive_set(
            {'storage-region': default_zone},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )
        config_manager.interactive_set(
            {'storage-account-name': ''},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )
        config_manager.interactive_set(
            {'storage-account-key': ''},
            secret_name=CONFIG_NAME,
            interactive=interactive
        )


def _get_cred_options():
    account_name = config_manager.get('storage-account-name', secret_name=CONFIG_NAME)
    account_key = config_manager.get('storage-account-key', secret_name=CONFIG_NAME)
    return f'--account-name {account_name} --account-key {account_key}'


def create_bucket(instance_id, region=None, exists_ok=False, dry_run=False):
    if not region:
        region = cluster_config_get('azure-default-location')

    assert region, 'No default region set for the cluster'

    bucket_exists = instance_id in list_azure_buckets()
    if not exists_ok and bucket_exists:
        raise Exception('Bucket for this instance already exists')

    if not dry_run and not bucket_exists:
        cred_options = _get_cred_options()
        az_check_output(f'storage container create -n {instance_id} {cred_options}')

    return {
        'BUCKET_NAME': instance_id,
        'STORAGE_ACCOUNT_NAME': config_manager.get('storage-account-name', secret_name=CONFIG_NAME),
        'STORAGE_ACCOUNT_KEY': config_manager.get('storage-account-key', secret_name=CONFIG_NAME)
    }


def delete_bucket(instance_id, dry_run=False):
    if instance_id not in list_azure_buckets():
        logs.warning(f'No bucket found for the instance "{instance_id}". Skipping.')
        return

    if dry_run:
        return

    logs.info(f'Removing Azure storage bucket for instance_id {instance_id}')

    cred_options = _get_cred_options()
    az_check_output(f'storage container delete -n {instance_id} {cred_options}')


def get_bucket(instance_id):
    if instance_id not in list_azure_buckets():
        logs.warning(f'No bucket found for the instance "{instance_id}" in this Azure storage account. Skipping.')
        return

    instance = kubectl.get(f'ckancloudckaninstance {instance_id}')
    bucket = instance['spec'].get('bucket').get(PROVIDER_ID)
    if not bucket:
        logs.warning('This instance does not have Azure bucket attached.')
        return

    return {
        'instance_id': instance_id,
        'bucket': bucket
    }


def list_buckets():
    """Returns list of buckets attached to CKAN Instances"""
    result = []
    for item in kubectl.get('ckancloudckaninstance').get('items', []):
        bucket = item['spec'].get('bucket', {}).get(PROVIDER_ID)
        if not bucket:
            continue
        result.append({
            'instance_id': item['spec']['id'],
            'bucket': bucket
        })

    return result


def list_azure_buckets():
    """Returns list of containers on Azure storage account"""
    cred_options = _get_cred_options()
    data = json.loads(az_check_output(f'storage container list {cred_options}'))

    return [x['name'] for x in data]



import click
import datetime
import traceback

from ckan_cloud_operator import logs

from . import manager


@click.group()
def gcloudsql():
    """Manage the gcloudsql centralized db proxy"""
    pass


@gcloudsql.command()
def port_forward():
    while True:
        start_time = datetime.datetime.now()
        try:
            manager.start_port_forward()
        except Exception:
            traceback.print_exc()
        end_time = datetime.datetime.now()
        if (end_time - start_time).total_seconds() < 10:
            logs.critical('DB Proxy failure')
            logs.exit_catastrophic_failure()
        else:
            logs.warning('Restarting the DB proxy')


@gcloudsql.command()
def initialize():
    manager.initialize()
    logs.exit_great_success()



from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs


DEFAULT_CHART_VALUES = {
    'chart-repo-name': 'stable',
    'chart-name': 'stable/jenkins',
}

DEFAULT_VALUES = {
    'Agent': {
        'Enabled': False,
    },
    'Master': {
        'ServiceType': 'ClusterIP',
    },
    'Persistence': {
        'Size': '100Gi',
        'StorageClass': 'cca-ckan'
    }
}


def pre_update_hook(instance_id, instance, res, sub_domain, root_domain, modify_spec_callback):
    logs.info('jenkins pre_update_hook', instance_id=instance_id, instance_spec=instance.get('spec'))
    chart_update_kwargs = {
        k: v for k, v in DEFAULT_CHART_VALUES.items()
        if not instance['spec'].get(k)
    }
    values_update_kwargs = {
        k: v for k, v in DEFAULT_VALUES.items()
        if not instance['spec'].get('values', {}).get(k)
    }
    logs.debug(chart_update_kwargs=chart_update_kwargs)
    logs.debug(values_update_kwargs=values_update_kwargs)
    instance['spec'].update(**chart_update_kwargs)
    instance['spec'].setdefault('values', {}).update(**values_update_kwargs)
    modify_spec_callback(lambda i: i.update(**chart_update_kwargs))
    modify_spec_callback(lambda i: i.setdefault('values', {}).update(**values_update_kwargs))


def pre_deploy_hook(instance_id, instance, deploy_kwargs):
    pass


def post_deploy_hook(instance_id, instance, deploy_kwargs):
    kubectl.apply(kubectl.get_deployment(
        name='jnlp-kube-prod-1',
        labels={},
        spec={
            'minReadySeconds': 15, 'progressDeadlineSeconds': 1200, 'replicas': 1,
            'revisionHistoryLimit': 10,
            'selector': {
                'matchLabels': {
                    'app': 'deployment-jenkins-jnlp-kube-prod-1'
                }
            },
            'strategy': {
                'type': 'Recreate'
            },
            'template': {
                'metadata': {
                    'labels': {
                        'app': 'deployment-jenkins-jnlp-kube-prod-1'
                    }
                },
                'spec': {
                    'containers': [
                        {
                            'command': ['bash', '/home/jenkins/ckan-cloud-operator/entrypoint-jnlp.sh'],
                            'env': [
                                {'name': 'CKAN_CLOUD_OPERATOR_SCRIPTS', 'value': '/usr/src/ckan-cloud-operator/scripts'},
                                {'name': 'CKAN_CLOUD_USER_NAME', 'value': 'jenkins-admin'},
                                {'name': 'HOME', 'value': '/home/jenkins/agent'}
                            ],
                            'envFrom': [
                                {'secretRef': {'name': 'jnlp-slave-kube-prod-1', 'optional': False}}
                            ],
                            'image': 'viderum/ckan-cloud-operator:jnlp-v0.2.7',
                            'imagePullPolicy': 'Always',
                            'name': 'jnlp-kube-prod-1',
                            'resources': {},
                            'volumeMounts': [
                                {'mountPath': '/etc/ckan-cloud', 'name': 'vol1'},
                                {'mountPath': '/home/jenkins/agent', 'name': 'workspace-volume'}
                            ],
                            'workingDir': '/home/jenkins/agent'
                        }
                    ],
                    'terminationGracePeriodSeconds': 120,
                    'volumes': [
                        {
                            'name': 'vol1',
                            'secret': {
                                'defaultMode': 511, 'optional': False, 'secretName': 'etc-ckan-cloud-jenkins-admin'
                            }
                        },
                        {
                            'emptyDir': {},
                            'name': 'workspace-volume'}
                    ]
                }
            }
        },
        namespace=instance_id,
    ))
    pass


def pre_delete_hook(instance_id, instance, delete_kwargs):
    pass


def post_delete_hook(instance_id, instance, delete_kwargs):
    pass


def get_backend_url(instance_id, instance, backend_url):
    release_name = instance['spec'].get('release-name')
    if release_name:
        return f'http://{release_name}-jenkins.{instance_id}:8080'
    else:
        return None


def get(instance_id, instance, res):
    res['ready'] = True
    app_pods_status = {}
    for pod in kubectl.get('pods', namespace=instance_id, required=True)['items']:
        app = pod['metadata']['labels'].get('app')
        if not app:
            app = 'unknown'
        item_status = kubectl.get_item_detailed_status(pod)
        if item_status.get('errors') and len(item_status['errors']) > 0:
            res['ready'] = False
        app_pods_status.setdefault(app, {})[pod['metadata']['name']] = item_status
    app_deployments_status = {}
    for deployment in kubectl.get('deployments', namespace=instance_id, required=True)['items']:
        app = deployment['metadata']['labels'].get('app')
        if not app:
            app = 'unknown'
        item_status = kubectl.get_item_detailed_status(deployment)
        if item_status.get('errors') and len(item_status['errors']) > 0:
            res['ready'] = False
        app_deployments_status.setdefault(app, {})[deployment['metadata']['name']] = item_status
    if 'jenkins' not in app_pods_status:
        res['ready'] = False
    res['app'] = {
        'pods': app_pods_status,
        'deployments': app_deployments_status
    }



import os
import subprocess
import yaml
import json

from dataflows import Flow, update_resource, printer, checkpoint, dump_to_path


def get_kube_items(operator, what, spec_fields=None, metadata_fields=None, status_fields=None):
    for item in yaml.load(subprocess.check_output(
        f'{operator} kubectl -- get {what} -o yaml', shell=True
    ))['items']:
        yield {
            **{k: v for k, v
               in item.get('spec', {}).items()
               if spec_fields is None or k in spec_fields},
            **{k: v for k, v
               in item.get('metadata', {}).items()
               if metadata_fields is None or k in metadata_fields},
            **{k: v for k, v
               in item.get('status', {}).items()
               if status_fields is None or k in status_fields}
        }



def cluster_info(operator):
    output = subprocess.check_output(f'{operator} cluster info', shell=True)
    yield from (cluster for cluster in yaml.load(output))


def ckan_cloud_instances(operator):

    def _processor(package):
        package.pkg.add_resource({
            'name': 'ckan-cloud-instances',
            'path': 'ckan-cloud-instances.csv',
            'schema': {
                'fields': [
                    {'name': n, 'type': 'string'} for n in
                    ['name', 'url', 'image', 'datastore', 'db', 'solr',
                     'storage', 'creationTimestamp', 'generation']
                ]
            }
        })
        yield package.pkg

        def get_normalized():
            items = get_kube_items(
                operator, 'ckancloudckaninstance',
                spec_fields=[
                    'ckanContainerSpec', 'ckanPodSpec', 'datastore', 'db',
                    'envvars', 'solrCloudCollection', 'storage'
                ],
                metadata_fields=[
                    'creationTimestamp', 'generation', 'name', 'namespace', 'resourceVersion', 'uid'
                ],
                status_fields=[]
            )
            for row in items:
                dbs = {}
                for db_type in ['datastore', 'db']:
                    dbs[db_type] = row[db_type]['name']
                    if row[db_type].get('dbPrefix'):
                        dbs[db_type] += f" ({row[db_type]['dbPrefix']})"
                solr = f"{row['solrCloudCollection']['name']} ({row['solrCloudCollection']['configName']})"
                instance_id = row['name']
                try:
                    url = subprocess.check_output(
                        f"{operator} deis-instance ckan exec {instance_id} -- -- "
                          f"bash -c 'env | grep CKAN_SITE_URL='",
                        shell=True
                    ).decode().split('=')[1]
                except Exception:
                    url = ''
                yield {
                    'name': row['name'],
                    'url': url.strip(),
                    'image': (
                        row['ckanContainerSpec']['image']
                        if row['ckanContainerSpec'].get('image')
                        else f"{row['ckanContainerSpec']['imageFromGitlab']} (imageFromGitlab)"
                    ),
                    'datastore': dbs['datastore'],
                    'db': dbs['db'],
                    'solr': solr,
                    'storage': row['storage']['path'],
                    'creationTimestamp': str(row['creationTimestamp']),
                    'generation': str(row['generation']),
                }
            for row in add_non_ckan_instances():
                yield row

        for resource in package:
            yield resource
        yield get_normalized()

    def add_non_ckan_instances():
        non_ckan_instances = os.environ.get('NON_CKAN_INSTANCES')
        if non_ckan_instances:
            non_ckan_instances = json.loads(non_ckan_instances)
            for instance in non_ckan_instances:
                instance_name = instance['name']
                get_route_args = instance['get-route-args']
                deployment = yaml.load(
                    subprocess.check_output(
                        f'{operator} kubectl -- -n {instance_name} get deployment {instance_name} -o yaml',
                        shell=True)
                )
                yield {
                    'name': instance_name,
                    'url': 'https://' + yaml.load(subprocess.check_output(
                        f'{operator} routers get-routes {get_route_args} --one',
                        shell=True
                    ))['frontend-hostname'],
                    'image': deployment['spec']['template']['spec']['containers'][0]['image'],
                    'datastore': '',
                    'db': '',
                    'solr': '',
                    'storage': '',
                    'creationTimestamp': str(deployment['metadata']['creationTimestamp']),
                    'generation': '',
                }

    return _processor


def main_flow(prefix, operator):
    return Flow(
        cluster_info(operator),
        update_resource(['res_1'], name='cluster-info', path='cluster-info.csv'),
        checkpoint(f'{prefix}-cluster-info'),
        ckan_cloud_instances(operator),
        update_resource(['res_2'], name='ckan-cloud-instances', path='ckan-cloud-instances.csv'),
    )


if __name__ == '__main__':
    prefix = os.environ['DATAPACKAGE_PREFIX']
    operator = os.environ.get('CKAN_CLOUD_OPERATOR_BIN', 'ckan-cloud-operator')
    Flow(
        main_flow(prefix, operator),
        printer(num_rows=1),
        dump_to_path(f'data/{prefix}/resources')
    ).process()



PROVIDER_SUBMODULE='apps'

APP_CRD_SINGULAR = 'app'
APP_CRD_PLURAL_SUFFIX = 'apps'
APP_CRD_KIND_SUFFIX = 'App'

APP_NAME_CRD_SINGULAR = 'appname'
APP_NAME_CRD_PLURAL_SUFFIX = 'appnames'
APP_NAME_CRD_KIND_SUFFIX = 'AppName'



PROVIDER_SUBMODULE = 'cluster'

# Kubernetes namespace used to store operator resources
# this namespace should only keep objects created and managed by the operator
OPERATOR_NAMESPACE = 'ckan-cloud'

# Kubernetes configmap name in the operator namespace, used to store global operator configurations
OPERATOR_CONFIGMAP = 'operator-conf'



import subprocess
import json
import os
import glob
import time

from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.infra import CkanInfra
from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl

from .constants import PROVIDER_SUBMODULE
from .solrcloud.constants import PROVIDER_ID as solrcloud_provider_id


def initialize(interactive=False, dry_run=False):
    ckan_infra = CkanInfra(required=False)
    solr_config = config_manager.interactive_set(
        {
            'self-hosted': True
        },
        secret_name='solr-config',
        interactive=interactive
    )
    if is_self_hosted(solr_config):
        initialize_self_hosted(interactive=interactive, dry_run=dry_run)
    else:
        config_manager.interactive_set(
            {
                'http-endpoint': ckan_infra.SOLR_HTTP_ENDPOINT,
            },
            secret_name='solr-config',
            interactive=interactive
        )
    config_manager.interactive_set(
        {
            'num-shards': ckan_infra.SOLR_NUM_SHARDS or '1',
            'replication-factor': ckan_infra.SOLR_REPLICATION_FACTOR or '3'
        },
        secret_name='solr-config',
        interactive=interactive
    )


def initialize_self_hosted(interactive=False, dry_run=False):
    get_provider(default=solrcloud_provider_id, verbose=True).initialize(interactive=interactive, dry_run=dry_run)
    zk_set_url_scheme()


def get_internal_http_endpoint():
    if is_self_hosted():
        return get_provider().get_internal_http_endpoint()
    else:
        return config_get('http-endpoint', required=True)


def is_self_hosted(config_vals=None):
    if config_vals:
        config_val = config_vals['self-hosted']
    else:
        config_val = config_get('self-hosted', required=True)
    return config_val == 'y'


def get_num_shards():
    return config_get('num-shards')


def get_replication_factor():
    return config_get('replication-factor')


def config_get(key, required=False):
    return config_manager.get(key, secret_name='solr-config', required=required)


def get_provider(default=None, verbose=False):
    from ckan_cloud_operator.providers import manager as providers_manager
    return providers_manager.get_provider(PROVIDER_SUBMODULE, default=default, verbose=verbose)


def start_zoonavigator_port_forward():
    get_provider().start_zoonavigator_port_forward()


def start_solrcloud_port_forward(suffix='sc-0'):
    get_provider().start_solrcloud_port_forward(suffix=suffix)


def delete_collection(collection_name):
    solr_curl(f'/admin/collections?action=DELETE&name={collection_name}', required=True)


def get_collection_status(collection_name):
    output = solr_curl(f'/{collection_name}/schema')
    if output == False:
        return {'ready': False,
                'collection_name': collection_name,
                'solr_http_endpoint': get_internal_http_endpoint()}
    else:
        def_ver, def_name = '2.8', 'ckan'
        res = {'schema': {'version': def_ver, 'name': def_name}}
        try:
            res = json.loads(output)
        except json.decoder.JSONDecodeError as e:
            logs.warning(
                f'Not able to decode response from SOLR. Using default values for schema version/name - {def_ver}/{def_name}\n SOLR response: \n{output}'
            )

        return {'ready': True,
                'collection_name': collection_name,
                'solr_http_endpoint': get_internal_http_endpoint(),
                'schemaVersion': res['schema']['version'],
                'schemaName': res['schema']['name']}


def create_collection(collection_name, config_name):
    logs.info(f'creating solrcloud collection {collection_name} using config {config_name}')
    replication_factor = get_replication_factor()
    num_shards = get_num_shards()
    output = solr_curl(f'/admin/collections?action=CREATE'
                       f'&name={collection_name}'
                       f'&collection.configName={config_name}'
                       f'&replicationFactor={replication_factor}'
                       f'&numShards={num_shards}', required=True)
    logs.info(output)


def solr_curl(path, required=False, debug=False):
    if is_self_hosted():
        return get_provider().solr_curl(path, required=required, debug=debug)
    else:
        http_endpoint = get_internal_http_endpoint()
        if debug:
            subprocess.check_call(f'curl \'{http_endpoint}{path}\'')
        else:
            exitcode, output = subprocess.getstatusoutput(f'curl -s -f \'{http_endpoint}{path}\'')
            if exitcode == 0:
                return output
            elif required:
                logs.critical(output)
                raise Exception(f'Failed to run solr curl: {http_endpoint}{path}')
            else:
                logs.warning(output)
                return False

def zk_set_url_scheme(scheme='http', timeout=300):
    pod_name = kubectl.get('pods', '-l', 'app=provider-solr-solrcloud-zk', required=True)['items'][0]['metadata']['name']
    try:
        kubectl.check_output('exec %s zkCli.sh set /clusterprops.json \'{"urlScheme":"%s"}\'' % (pod_name, scheme))
    except Exception as e:
        print('Failed to connect ZooKeeper, retrying in 60 seconds')
        time.sleep(60)
        if timeout < 0:
            raise e
        zk_set_url_scheme(scheme=scheme, timeout=timeout-60)


def zk_list_configs():
    pod_name = kubectl.get('pods', '-l', 'app=provider-solr-solrcloud-zk', required=True)['items'][0]['metadata']['name']
    lines = list(kubectl.check_output(f'exec {pod_name} zkCli.sh ls /configs').decode().splitlines())[5:]
    if len(lines) == 1:
        return [name.strip() for name in lines[0][1:-1].split(',')]
    else:
        return []


def zk_list_config_files(config_name, config_files, base_path=''):
    path = f'/configs/{config_name}{base_path}'
    # print(f'path={path}')
    pod_name = kubectl.get('pods', '-l', 'app=provider-solr-solrcloud-zk', required=True)['items'][0]['metadata']['name']
    lines = list(kubectl.check_output(f'exec {pod_name} zkCli.sh ls {path}').decode().splitlines())[5:]
    # print(f'lines={lines}')
    assert len(lines) == 1
    num_files = 0
    for name in lines[0][1:-1].split(','):
        name = name.strip()
        if not name: continue
        # print(f'name={name}')
        if zk_list_config_files(config_name, config_files, base_path=f'{base_path}/{name}') == 0:
            config_files.append(f'{base_path}/{name}')
            num_files += 1
    return num_files


def zk_get_config_file(config_name, config_file, output_filename):
    path = f'/configs/{config_name}{config_file}'
    # print(f'path={path}')
    pod_name = kubectl.get('pods', '-l', 'app=provider-solr-solrcloud-zk', required=True)['items'][0]['metadata']['name']
    lines = list(kubectl.check_output(f'exec {pod_name} zkCli.sh get {path} 2>/dev/null').decode().splitlines())[5:]
    assert len(lines) > 0
    os.makedirs(os.path.dirname(output_filename), exist_ok=True)
    with open(output_filename, 'w') as f:
        f.write('\n'.join(lines))


def zk_put_configs(configs_dir):
    def retry_if_fails(command, max_retries=15):
        if max_retries < 0:
            return
        try:
            kubectl.check_output(command)
        except:
            time.sleep(5)
            retry_if_fails(command, max_retries=max_retries-1)

    pod_name = kubectl.get('pods', '-l', 'app=provider-solr-solrcloud-zk', required=True)['items'][0]['metadata']['name']
    logs.info(f'using pod {pod_name}')
    for input_filename in glob.glob(f'{configs_dir}/**/*', recursive=True):
        if not os.path.isfile(input_filename): continue
        output_filename = '/configs' + input_filename.replace(configs_dir, '')
        logs.info(f'{input_filename} --> {output_filename}')
        output_filepath = ''
        for output_filepart in output_filename.split('/')[:-1]:
            output_filepart = output_filepart.strip()
            if not output_filepart:
                continue
            output_filepath += f'/{output_filepart}'
            logs.info(f'create {output_filepath} null')
            retry_if_fails(f'exec {pod_name} zkCli.sh create {output_filepath} null')
        logs.info(f'copy {output_filename}')
        retry_if_fails(f'cp {input_filename} {pod_name}:/tmp/zk_input')
        logs.info(f'create {output_filename}')
        retry_if_fails(f"exec {pod_name} -- /bin/bash -c '/usr/bin/zkCli.sh create {output_filename} \"$(cat /tmp/zk_input)\"'")



import subprocess
import functools
from ckan_cloud_operator.infra import CkanInfra


@functools.lru_cache(maxsize=1)
def activate():
    subprocess.check_call('ckan-cloud-operator activate-gcloud-auth', shell=True)
    return True


def check_output(cmd, project=None, with_activate=True, ckan_infra=None, gsutil=False):
    if with_activate: activate()
    if not ckan_infra:
        ckan_infra = CkanInfra()
    if not project:
        project = ckan_infra.GCLOUD_AUTH_PROJECT
    compute_zone = ckan_infra.GCLOUD_COMPUTE_ZONE
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.check_output(f'{bin} {cmd}', shell=True)


def check_call(cmd, project=None, with_activate=True, gsutil=False, ckan_infra=None):
    if with_activate: activate()
    if not ckan_infra:
        ckan_infra = CkanInfra()
    if not project:
        project = ckan_infra.GCLOUD_AUTH_PROJECT
        compute_zone = ckan_infra.GCLOUD_COMPUTE_ZONE
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.check_call(f'{bin} {cmd}', shell=True)


def getstatusoutput(cmd, project=None, with_activate=True, gsutil=False, ckan_infra=None):
    if with_activate: activate()
    if not ckan_infra:
        ckan_infra = CkanInfra()
    if not project:
        project = ckan_infra.GCLOUD_AUTH_PROJECT
        compute_zone = ckan_infra.GCLOUD_COMPUTE_ZONE
    bin = 'gsutil' if gsutil else f'CLOUDSDK_COMPUTE_ZONE={compute_zone} gcloud --project={project}'
    return subprocess.getstatusoutput(f'{bin} {cmd}')



import json
import unittest
from unittest.mock import patch

from ckan_cloud_operator.providers.solr import manager


class SolrManagerTestCase(unittest.TestCase):
    @patch('ckan_cloud_operator.providers.solr.manager.get_replication_factor')
    @patch('ckan_cloud_operator.providers.solr.manager.get_num_shards')
    @patch('ckan_cloud_operator.providers.solr.manager.solr_curl')
    def test_create_collection(self, solr_curl, get_num_shards, get_replication_factor):
        get_replication_factor.return_value = 2
        get_num_shards.return_value = 1
        solr_curl.return_value = ''

        manager.create_collection('montreal', 'config_27_default')
        solr_curl.assert_called_once_with('/admin/collections?action=CREATE&name=montreal&collection.configName=config_27_default&replicationFactor=2&numShards=1', required=True)

    @patch('ckan_cloud_operator.providers.solr.manager.solr_curl')
    @patch('ckan_cloud_operator.providers.solr.manager.get_internal_http_endpoint')
    def test_get_collection_status(self, http_endpoint, solr_curl):
        http_endpoint.return_value = '192.168.0.101'
        solr_curl.return_value = json.dumps({
            'schema': {
                'version': '27',
                'name': 'ckan_27_default'
            }
        })
        expected_status = {
            'ready': True,
            'collection_name': 'montreal',
            'solr_http_endpoint': '192.168.0.101',
            'schemaVersion': '27',
            'schemaName': 'ckan_27_default'
        }
        self.assertEqual(manager.get_collection_status('montreal'), expected_status)

    @patch('ckan_cloud_operator.providers.solr.manager.solr_curl')
    @patch('ckan_cloud_operator.providers.solr.manager.get_internal_http_endpoint')
    def test_get_collection_status_not_ready(self, http_endpoint, solr_curl):
        http_endpoint.return_value = '192.168.0.101'
        solr_curl.return_value = False
        expected_status = {
            'ready': False,
            'collection_name': 'montreal',
            'solr_http_endpoint': '192.168.0.101'
        }
        self.assertEqual(manager.get_collection_status('montreal'), expected_status)



import click

from ckan_cloud_operator import logs
from . import manager


@click.group('azuresql')
def azuresql():
    """Manage Azure SQL instances for centralized DB services"""
    pass


@azuresql.command()
@click.option('--interactive', is_flag=True)
@click.option('--db-prefix')
def initialize(interactive, db_prefix):
    """Enable Azure SQL as the main DB provider"""
    manager.initialize(interactive=interactive, db_prefix=db_prefix)
    logs.exit_great_success()



import os
import tempfile
from ruamel import yaml
from ckan_cloud_operator import logs


# Functions to support automation scripts (usually under scripts/ directory)


def check_file_based_approval_code(approve_code, expected_content, delete=True):
    if approve_code:
        approval_filename = f'data/approval_files/{approve_code}'
        assert os.path.isfile(approval_filename), 'invalid approval code'
        with open(approval_filename) as f:
            assert yaml.safe_load(f) == expected_content, 'invalid approval code or expected content was changed'
        if delete:
            os.unlink(approval_filename)
        return True
    else:
        return False


def create_file_based_approval_code(content):
    os.makedirs('data/approval_files', exist_ok=True)
    with tempfile.NamedTemporaryFile('w', prefix='cca', dir='data/approval_files', delete=False) as f:
        logs.yaml_dump(content, f)
        return os.path.relpath(f.name, os.path.join(os.path.curdir, 'data/approval_files'))


def get_env_vars(*args):
    return [os.environ.get(e, '') for e in args]



#### standard provider code ####

# import the correct PROVIDER_SUBMODULE and PROVIDER_ID constants for your provider
from .constants import PROVIDER_ID
from ..constants import PROVIDER_SUBMODULE

# define common provider functions based on the constants
from ckan_cloud_operator.providers import manager as providers_manager
def _get_resource_name(suffix=None): return providers_manager.get_resource_name(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _get_resource_labels(for_deployment=False): return providers_manager.get_resource_labels(PROVIDER_SUBMODULE, PROVIDER_ID, for_deployment=for_deployment)
def _get_resource_annotations(suffix=None): return providers_manager.get_resource_annotations(PROVIDER_SUBMODULE, PROVIDER_ID, suffix=suffix)
def _set_provider(): providers_manager.set_provider(PROVIDER_SUBMODULE, PROVIDER_ID)
def _config_set(key=None, value=None, values=None, namespace=None, is_secret=False, suffix=None): providers_manager.config_set(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, value=value, values=values, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_get(key=None, default=None, required=False, namespace=None, is_secret=False, suffix=None): return providers_manager.config_get(PROVIDER_SUBMODULE, PROVIDER_ID, key=key, default=default, required=required, namespace=namespace, is_secret=is_secret, suffix=suffix)
def _config_interactive_set(default_values, namespace=None, is_secret=False, suffix=None, from_file=False): providers_manager.config_interactive_set(PROVIDER_SUBMODULE, PROVIDER_ID, default_values, namespace, is_secret, suffix, from_file)

################################
# custom provider code starts here
#

import os
import yaml
import binascii

from ckan_cloud_operator import logs
from ckan_cloud_operator import kubectl



def initialize(interactive=False):
    _set_provider()
    if interactive:
        print('\nUsing Minikube\n')
    print(yaml.dump(get_info(), default_flow_style=False))
    _create_storage_classes()


def get_info(debug=False):
    return {}

def _create_storage_classes():
    for sc in ('cca-storage', 'cca-ckan'):
        kubectl.apply({
            'apiVersion': 'storage.k8s.io/v1', 'kind': 'StorageClass',
            'metadata': {
                'annotations': {'storageclass.kubernetes.io/is-default-class': 'true'},
                'labels': {'addonmanager.kubernetes.io/mode': 'EnsureExists'},
                'name': sc,
            },
            'provisioner': 'k8s.io/minikube-hostpath',
            'reclaimPolicy': 'Delete',
            'volumeBindingMode': 'Immediate'
        })



def create_volume(disk_size_gb, labels, use_existing_disk_name=None, zone=None):
    disk_id = use_existing_disk_name or 'cc' + _generate_password(12)
    if use_existing_disk_name:
        logs.info(f'using existing persistent disk {disk_id}')
    else:
        logs.info(f'creating persistent disk {disk_id} with size {disk_size_gb}')
    kubectl.apply({
        'apiVersion': 'v1', 'kind': 'PersistentVolume',
        'metadata': {'name': disk_id, 'namespace': 'ckan-cloud'},
        'spec': {
            'capacity': {'storage': f'{disk_size_gb}G'},
            'accessModes': ['ReadWriteOnce'],
            'hostPath': {'path': '/data/' + disk_id},
            'storageClassName': '',
        }
    })
    kubectl.apply({
        'apiVersion': 'v1', 'kind': 'PersistentVolumeClaim',
        'metadata': {'name': disk_id, 'namespace': 'ckan-cloud'},
        'spec': {
            'volumeName': disk_id,
            'accessModes': ['ReadWriteOnce'],
            'resources': {'requests': {'storage': f'{disk_size_gb}G'}},
            'storageClassName': '',
        }
    })
    return {'persistentVolumeClaim': {'claimName': disk_id}}


def _generate_password(l):
    return binascii.hexlify(os.urandom(l)).decode()



import requests
from urllib.parse import urljoin

from ...env import manager as env_manager


class Drone(object):

    def initialize(self, force_prompt=False):
        self.conf = env_manager._read_yaml().get('cicd', {})
        self.server_url = self.conf.get('serverUrl')
        self.api_token = self.conf.get('token')
        self.org = self.conf.get('organization')
        self.repo = self.conf.get('repo')
        if force_prompt or not self.conf:
            self.server_url = input('Please enter Drone Server URL: ')
            self.api_token = input(f'Please enter Drone API Token. See {self. server_url} acount: ')
            self.org = input('Please enter Github organization name: ')
            self.repo = input('Please enter Github repository name: ')
            self.conf['cicd'] = {
                 'serverUrl': self.server_url,
                 'token': self.api_token,
                 'organization': self.org,
                 'repo': self.repo
            }
            env_manager._write_yaml(self.conf)
        self.api_url = urljoin(self.server_url, '/'.join(['api/repos', f'{self.org}/{self.repo}', 'builds/']))
        self.header = {'Authorization': f'Bearer {self.api_token}'}
        print('CCO is configured for Drone ')


    def builds_list(self):
        resp = requests.get(self.api_url, headers=self.header)
        if not _check_for_200(resp):
            return []
        return resp.json()


    def builds_info(self, branch='develop'):
        build_number = self.get_build_number(branch=branch)
        if build_number is None:
            print(f'Build {build_number} not found')
            return None
        url = urljoin(self.api_url, build_number)
        builds_info_resp = requests.get(url, headers=self.header)
        if not _check_for_200(builds_info_resp):
            return {}
        return builds_info_resp.json()


    def builds_logs(self, branch='develop'):
        build_info = self.builds_info(branch=branch)
        if build_info is None:
            return None
        stages = build_info.get('stages', [])
        for stage in stages:
            steps = stage.get('steps', [])
            stage_name = stage.get('name')
            for step in steps:
                step_name = step.get('name')
                print(f'--- Build Stage: {stage_name} | Builds Step: {step_name}')
                url = urljoin(self.api_url, '/'.join([
                    self.build_number,
                    'logs',
                    str(stage.get('number')),
                    str(step.get('number'))])
                )
                log_resp = requests.get(url, headers=self.header)
                log_list = log_resp.json()
                for _log in log_list:
                    print(_log.get('out').rstrip('\n'))
        print(f'Showing logs for "{branch}" Branch')


    def get_build_number(self, branch='develop'):
        for build in self.builds_list():
            if build.get('source') == branch:
                self.build_number = str(build.get('number'))
                return str(build.get('number'))
        print(f'No build found for branch {branch}')
        return None


def _check_for_200(resp):
    if resp.status_code == 200:
        return True
    if resp.status_code == 401:
        print("Seems like you are not authorized")
        print("Please run `cco ckan deployment drone initialize --force-update` and rerun")
        return False
    return False



import hashlib

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs

from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.labels import manager as labels_manager
from ckan_cloud_operator.annotations import manager as annotations_manager


def initialize(log_kwargs=None):
    values = {
        'crd-group': 'stable.viderum.com',
        'crd-prefix': 'CkanCloud'
    }
    logs.info(f'Setting default crds module configurations', **(log_kwargs or {}), **values)
    config_manager.set(values=values)


def get_crd_group():
    return config_manager.get('crd-group')


def get_crd_prefix():
    return config_manager.get('crd-prefix')


def get(singular, *args, name=None, required=True, get_cmd='get', **kwargs):
    """Run kubectl.get for the given crd singular value and optional get args / kwargs"""
    crd_prefix = get_crd_prefix()
    _, kind_suffix = _get_plural_kind_suffix(singular)
    if name:
        args = [get_resource_name(singular, name), *args]
    return kubectl.get(f'{crd_prefix}{kind_suffix}', *args, required=required, get_cmd=get_cmd, **kwargs)


def edit(singular, *edit_args, name=None, **edit_kwargs):
    """Run kubectl.get for the given crd singular value and optional get args / kwargs"""
    crd_prefix = get_crd_prefix()
    _, kind_suffix = _get_plural_kind_suffix(singular)
    what = f'{crd_prefix}{kind_suffix}'
    if name:
        what += '/' + get_resource_name(singular, name)
    return kubectl.edit(what, *edit_args, **edit_kwargs)


def delete(singular, name):
    config_delete(singular, name, by_labels=True)
    crd_prefix = get_crd_prefix()
    _, kind_suffix = _get_plural_kind_suffix(singular)
    name = get_resource_name(singular, name)
    kubectl.check_call(f'delete --ignore-not-found {crd_prefix}{kind_suffix} {name}')


def install_crd(singular, plural_suffix, kind_suffix, hash_names=False):
    logs.info(f'Installing operator crd: {singular}, {plural_suffix}, {kind_suffix}')
    _set_plural_kind_suffix(singular, plural_suffix, kind_suffix, hash_names=hash_names)
    label_prefix = labels_manager.get_label_prefix().replace('-', '')
    kubectl.install_crd(
        f'{label_prefix}{plural_suffix}',
        f'{label_prefix}{singular}',
        get_resource_kind(singular)
    )


def list_crds(full=False, debug=False):
    for config_key in config_manager.get():
        if config_key.startswith('installed-crd-'):
            singular = config_key.replace('installed-crd-', '')
            yield get_crd(singular, full=full, debug=debug)


def get_crd(singular, full=True, debug=False):
    if debug: full = True
    plural_suffix, kind_suffix = _get_plural_kind_suffix(singular)
    label_prefix = labels_manager.get_label_prefix().replace('-', '')
    data = {'singular': f'{label_prefix}{singular}',
            'plural': f'{label_prefix}{plural_suffix}',
            'kind': get_resource_kind(singular),
            'singular-suffix': singular}
    if full:
        data['resources'] = []
        items = get(singular)
        if items:
            for resource in items.get('items', []):
                resource_data = {'name': resource.get('metadata', {}).get('name')}
                if debug:
                    resource_data['labels'] = resource.get('metadata', {}).get('labels')
                    resource_data['spec'] = resource.get('spec')
                data['resources'].append(resource_data)
    return data


def list_resources(singular=None, name=None):
    if singular:
        for item in get(singular).get('items', []):
            resource = {'name': item.get('metadata', {}).get('name')}
            yield resource
    else:
        for crd in list_crds(full=False, debug=False):
            for resource in list_resources(crd['singular-suffix'], name):
                resource['kind'] = crd['kind']
                yield resource


def get_resource(singular, name, extra_label_suffixes=None, **kwargs):
    crd_group = get_crd_group()
    _, kind_suffix = _get_plural_kind_suffix(singular)
    resource = kubectl.get_resource(
        f'{crd_group}/v1',
        get_resource_kind(singular),
        get_resource_name(singular, name),
        get_resource_labels(singular, name, extra_label_suffixes=extra_label_suffixes)
    )
    resource.update(**kwargs)
    return resource


def get_resource_name(singular, name, allow_hash_names=True):
    # print(f'get_resource_name: {singular} {name}')
    resource_name = labels_manager.get_resource_name(get_resource_suffix(singular, name))
    if _get_hash_names(singular) and allow_hash_names:
        resource_name = 'cc' + hashlib.blake2b(name.encode(), digest_size=16).hexdigest()
    # print(f'output resource name: {resource_name}')
    return resource_name


def get_resource_kind(singular):
    crd_prefix = get_crd_prefix()
    _, kind_suffix = _get_plural_kind_suffix(singular)
    return f'{crd_prefix}{kind_suffix}'


def get_resource_suffix(singular, name):
    parts = [singular]
    if name:
        parts.append(name)
    return '-'.join(parts)


def get_resource_labels(singular, name, extra_label_suffixes=None, for_deployment=False):
    label_suffixes = {
        f'crd-{singular}-name': name,
    }
    if extra_label_suffixes:
        label_suffixes.update(**extra_label_suffixes)
    extra_labels = {'app': get_deployment_app_label(singular, name)} if for_deployment else {}
    return labels_manager.get_resource_labels(label_suffixes, extra_labels=extra_labels)


def get_deployment_app_label(singular, name):
    return get_resource_suffix(singular, name)


def get_resource_annotations(singular, name):
    return annotations_manager.get_global_annotations()


def config_set(singular, name, key=None, value=None, values=None, namespace=None, is_secret=False):
    """store key/values in a secret or configmap"""
    resource_name = get_resource_name(singular, name)
    config_manager.set(
        key=key,
        value=value,
        values=values,
        secret_name=resource_name if is_secret else None,
        configmap_name=None if is_secret else resource_name,
        namespace=namespace,
        extra_operator_labels=_get_label_suffixes(singular, name)
    )


def config_get(singular, name, key=None, default=None, required=False, namespace=None, is_secret=False):
    resource_name = get_resource_name(singular, name)
    return config_manager.get(
        key=key,
        default=default,
        secret_name=resource_name if is_secret else None,
        configmap_name=None if is_secret else resource_name,
        required=required,
        namespace=namespace
    )


def config_delete(singular, name, namespace=None, is_secret=False, exists_ok=False, by_labels=False):
    resource_name = get_resource_name(singular, name)
    if by_labels:
        config_manager.delete_by_extra_operator_labels(_get_label_suffixes(singular, name))
    else:
        config_manager.delete(
            secret_name=resource_name if is_secret else None,
            configmap_name=None if is_secret else resource_name,
            namespace=namespace,
            exists_ok=exists_ok
        )


def config_interactive_set(singular, name, default_values=None, namespace=None, is_secret=False, from_file=False):
    resource_name = get_resource_name(singular, name)
    config_manager.interactive_set(
        default_values,
        secret_name=resource_name if is_secret else None,
        configmap_name=None if is_secret else resource_name,
        namespace=namespace,
        from_file=from_file
    )


def _get_plural_kind_suffix(singular):
    parts = config_manager.get(f'installed-crd-{singular}', required=True).split(',')
    plural_suffix = parts[0]
    kind_suffix = parts[1]
    return plural_suffix, kind_suffix


def _get_hash_names(singular):
    parts = config_manager.get(f'installed-crd-{singular}', required=True).split(',')
    return len(parts) > 2 and parts[2] == 'y'


def _set_plural_kind_suffix(singular, plural_suffix, kind_suffix, hash_names=False):
    hash_names = 'y' if hash_names else 'n'
    config_manager.set(f'installed-crd-{singular}', f'{plural_suffix},{kind_suffix},{hash_names}')


def _get_label_suffixes(singular, name):
    return {
        f'crd-{singular}-name': name,
        f'crd-singular': singular,
        f'crd-name': name,
    }



import click
import yaml

from ckan_cloud_operator import logs

from . import manager


@click.group()
def proxy():
    """Manage SOLR proxy for centralized unauthenticated access"""
    pass


@proxy.command()
def initialize():
    manager.deploy()
    logs.exit_great_success()



import base64
import os
import yaml
from sys import stdout

from ckan_cloud_operator import kubectl
from ckan_cloud_operator import logs

from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.labels import manager as labels_manager


__CACHED_VALUES = {}


def get(key=None, default=None, secret_name=None, configmap_name=None, namespace=None, required=False, template=None):
    cache_key = _get_cache_key(secret_name, configmap_name, namespace)
    if cache_key not in __CACHED_VALUES:
        __CACHED_VALUES[cache_key] = _fetch(cache_key)
    if key:
        value = __CACHED_VALUES[cache_key].get(key, default)
    else:
        value = __CACHED_VALUES[cache_key]
    assert value or not required, f'config value is required for {cache_key}:{key}'
    if template:
        if key:
            return template.format(**{key: value})
        else:
            return template.format(**value)
    else:
        return value


def set(key=None, value=None, values=None, secret_name=None, configmap_name=None, namespace=None, extra_operator_labels=None,
        from_file=False, dry_run=False):
    log_kwargs = {'func': 'config/set', 'secret': secret_name, 'configmap': configmap_name, 'namespace': namespace}
    cache_key = _get_cache_key(secret_name, configmap_name, namespace)
    if secret_name is None and configmap_name is None and namespace is None:
        # hack to set label-prefix to bootstrap the environment
        if key == 'label-prefix' and not values:
            label_prefix = value
        elif not key and 'label-prefix' in values:
            label_prefix = values['label-prefix']
        else:
            label_prefix = None
        if label_prefix:
            __CACHED_VALUES.setdefault(cache_key, {})['label-prefix'] = label_prefix
    logs.debug('start', **log_kwargs)
    if from_file:
        assert key and value and not values
        with open(value) as f:
            value = f.read()
        values = {key: value}
    elif key or value:
        assert key and value and not values, 'Invalid arguments: must specify both key and value args and not specify values arg'
        values = {key: value}
    assert values, 'Invalid arguments: no values to save'
    return _save(cache_key, values, extra_operator_labels, dry_run=dry_run)


def delete_key(key, secret_name=None, namespace=None):
    kubectl.apply({
        'apiVersion': 'v1',
        'kind': 'Secret',
        'metadata': {
            'name': secret_name,
            'namespace': namespace
        },
        'type': 'Opaque',
        'data': {
            k: base64.b64encode(v.encode()).decode()
            for k, v
            in kubectl.decode_secret(kubectl.get('secret', secret_name, namespace=namespace)).items()
            if k != key and v
        }
    })


def delete(secret_name=None, configmap_name=None, namespace=None, exists_ok=False):
    cache_key = _get_cache_key(secret_name, configmap_name, namespace)
    _delete(cache_key, exists_ok)


def delete_by_extra_operator_labels(extra_operator_labels):
    labels = labels_manager.get_resource_labels(extra_operator_labels)
    labels_manager.delete_by_labels(labels, kinds=['configmap', 'secret'])


def list_configs(namespace=None, full=False, show_secrets=False):
    label_prefix = labels_manager.get_label_prefix()
    if not namespace: namespace = cluster_manager.get_operator_namespace_name()
    what = 'configmaps'
    if show_secrets:
        what += ',secrets'
    configs = kubectl.get(what, '-l', f'{label_prefix}/operator-config-namespace={namespace}', required=False)
    if configs:
        for config in configs.get('items', []):
            kind = config['kind']
            name = config.get('metadata', {}).get('name')
            data = {'kind': config['kind'],
                    'name': config.get('metadata', {}).get('name')}
            if full:
                if name:
                    data['values'] = get(
                        secret_name=name if kind == 'Secret' else None,
                        configmap_name=name if kind == 'ConfigMap' else None,
                        namespace=namespace,
                        required=False
                    )
                else:
                    data['values'] = None
            yield data

def get_preset_answer(namespace, configmap_name, secret_name, key, default=None):
    interactive_file = os.environ.get('CCO_INTERACTIVE_CI')
    if not interactive_file:
        return
    answers = yaml.load(open(interactive_file))
    section = '?'
    subsection = '?'
    namespace = namespace or 'default'
    try:
        if configmap_name:
            section = 'config'
            subsection = configmap_name
        else:
            section = 'secrets'
            subsection = secret_name
        return answers[namespace][section][subsection][key]
    except:
        if default is not None:
            return default
        logs.error(f'Failed to find in interactive file value for {namespace}.{section}.{subsection}.{key}')
        raise


def interactive_set(default_values, secret_name=None, configmap_name=None, namespace=None, from_file=False,
                    extra_operator_labels=None, interactive=True):
    log_kwargs = {'func': 'config/interactive_set', 'secret': secret_name, 'configmap': configmap_name, 'namespace': namespace}
    logs.debug('start', **log_kwargs)
    set_values = {}
    for key, default_value in default_values.items():
        saved_value = get(key, secret_name=secret_name, configmap_name=configmap_name, namespace=namespace)
        preset_value = get_preset_answer(namespace, configmap_name, secret_name, key, default=default_value)
        if preset_value:
            set_values[key] = preset_value
        elif interactive and stdout.isatty():
            if saved_value:
                if from_file:
                    msg = ', leave empty to use the saved value'
                else:
                    msg = f', leave empty to use the saved value: {saved_value}'
                    default_value = saved_value
            elif default_value is not None:
                assert not from_file
                msg = f', leave empty to use the default value: {default_value}'
            else:
                msg = ' (required)'
            if from_file:
                print(f'Enter the path to a file containing the value for {key}{msg}')
                source_path = input(f'{key} path: ')
                if source_path:
                    with open(source_path) as f:
                        set_values[key] = f.read()
                elif saved_value:
                    set_values[key] = saved_value
                else:
                    raise Exception('file path is required')
            else:
                if default_value in [True, False]:
                    print(f'Enter a boolean value for {key}{msg}')
                    entered_value = input(f'{key} [y/n]: ')
                    bool_value = default_value if entered_value == '' else (entered_value == 'y')
                    set_values[key] = 'y' if bool_value else 'n'
                else:
                    print(f'Enter a value for {key}{msg}')
                    entered_value = input(f'{key}: ')
                    set_values[key] = str(entered_value or default_value)
        else:
            set_values[key] = saved_value if saved_value is not None else default_value
    logs.debug('set', **log_kwargs)
    return set(values=set_values, secret_name=secret_name, configmap_name=configmap_name, namespace=namespace, extra_operator_labels=extra_operator_labels)


def _fetch(cache_key):
    config_type, namespace, config_name = _parse_cache_key(cache_key)
    fetch_func = {
        'secret': lambda: _fetch_secret(config_name, namespace),
        'configmap': lambda: _fetch_configmap(config_name, namespace)
    }.get(config_type)
    assert fetch_func, f'Invalid config type: {config_type}'
    return fetch_func() or {}


def _fetch_secret(secret_name, namespace):
    secret = kubectl.get(f'secret {secret_name}', required=False, namespace=namespace)
    return kubectl.decode_secret(secret) if secret else None


def _fetch_configmap(configmap_name, namespace):
    configmap = kubectl.get(f'configmap {configmap_name}', required=False, namespace=namespace)
    return configmap['data'] if configmap else None


def _save(cache_key, values, extra_operator_labels, dry_run=False):
    config_type, namespace, config_name = _parse_cache_key(cache_key)
    save_func = {
        'secret': lambda: _save_secret(values, config_name, namespace, extra_operator_labels, dry_run=dry_run),
        'configmap': lambda: _save_configmap(values, config_name, namespace, extra_operator_labels, dry_run=dry_run),
    }.get(config_type)
    assert save_func, f'Invalid config type: {config_type}'
    __CACHED_VALUES[cache_key] = res = save_func()
    return res


def _save_secret(values, secret_name, namespace, extra_operator_labels, dry_run=False):
    return kubectl.update_secret(
        secret_name,
        values,
        namespace=namespace,
        labels=_get_labels(secret_name=secret_name, namespace=namespace, extra_operator_labels=extra_operator_labels),
        dry_run=dry_run
    )


def _save_configmap(values, configmap_name, namespace, extra_operator_labels, dry_run=False):
    return kubectl.update_configmap(
        configmap_name, values, namespace=namespace,
        labels=_get_labels(configmap_name=configmap_name, namespace=namespace, extra_operator_labels=extra_operator_labels),
        dry_run=dry_run
    )


def _delete(cache_key, exists_ok=False):
    config_type, namespace, config_name = _parse_cache_key(cache_key)
    assert config_type in ['configmap', 'secret'], f'Invalid config type: {config_type}'
    ignore_not_found = ' --ignore-not-found' if exists_ok else ''
    kubectl.check_call(f'delete{ignore_not_found} {config_type} {config_name}')


def _get_labels(cache_key=None, secret_name=None, configmap_name=None, namespace=None, extra_operator_labels=None):
    if cache_key:
        assert not secret_name and not configmap_name and not namespace
    else:
        cache_key = _get_cache_key(secret_name, configmap_name, namespace)
    config_type, namespace, config_name = _parse_cache_key(cache_key)
    operator_labels = {
        f'operator-config-{config_type}': config_name,
        'operator-config-namespace': namespace
    }
    if extra_operator_labels:
        operator_labels.update(**extra_operator_labels)
    return labels_manager.get_resource_labels(operator_labels)


def _get_cache_key(secret_name, configmap_name, namespace):
    if secret_name:
        assert not configmap_name, f'Invalid arguments: cannot specify both secret_name and configmap_name: {secret_name}, {configmap_name}'
    configmap_name, namespace = cluster_manager.get_operator_configmap_namespace_defaults(configmap_name, namespace)
    return f'secret:{namespace}:{secret_name}' if secret_name else f'configmap:{namespace}:{configmap_name}'


def _parse_cache_key(cache_key):
    config_type, namespace, config_name = cache_key.split(':')
    return config_type, namespace, config_name



import yaml
import time
import datetime
import binascii
import os
import traceback
import subprocess
import sys

from ckan_cloud_operator import kubectl, logs, yaml_config
from ckan_cloud_operator.config import manager as config_manager
from ckan_cloud_operator.crds import manager as crds_manager
from ckan_cloud_operator.labels import manager as labels_manager
from ckan_cloud_operator.providers.cluster import manager as cluster_manager
from ckan_cloud_operator.providers.storage.constants import CONFIG_NAME
from ckan_cloud_operator.providers.storage.manager import get_provider_id as get_storage_provider_id
from ckan_cloud_operator.routers import manager as routers_manager

from ..constants import INSTANCE_NAME_CRD_SINGULAR
from ..constants import INSTANCE_CRD_SINGULAR
from ..deployment import manager as deployment_manager


import yaml
from pathlib import Path


def list_environments():
    if _cco_is_configured() and _get_config():
        cco_config = _get_config()
        for environment in cco_config.get('environments', []):
            prefix = '>' if environment.get('active', False) else '-'
            print(prefix, environment.get('name'))
    else:
        print('Ckan Cloud Operator is no yet configured. Please run cco add << environment >>')

def add_environment(environment, **kwargs):
    cco_config = {}
    if _cco_is_configured() and _get_config():
        cco_config = _get_config()
        new_envs = [_env for _env in cco_config.get('environments') or [] if environment.lower() != _env.get('name', '').lower()]
        cco_config['environments'] = new_envs

    if kwargs['cloud_provider'] == 'azure':
        _run_az_command_or_login(
            f'az account set --subscription {kwargs["subscription"]}',
            add_environment,
            kwargs
        )
        kwargs['logged_in'] = True
        _run_az_command_or_login(
            f'az aks get-credentials --resource-group {kwargs["resource_group"]} --name {kwargs["cluster_name"]}',
            add_environment,
            kwargs
        )
        envs = cco_config.get('environments', []) or []
        for _env in envs:
            _env['active'] = False
        kwargs['active'] = True
        kwargs['name'] = environment
        envs.append(kwargs)
        cco_config['environments'] = envs
        # update config file
        _mkconfdir()
        _write_yaml(cco_config)
        print(f'{kwargs["cluster_name"]} has been successfully configured for "{environment}" environment')
    elif kwargs['cloud_provider'] == 'gcloud':
        #TODO
        pass
    elif kwargs['cloud_provider'] == 'aws':
        #TODO
        pass
    elif kwargs['cloud_provider'] == 'minio':
        #TODO
        pass
    else:
        logs.warning(f'Cloud Provider "{cloud_provider}" is not supported')

def update_environment(environment, **kwargs):
    updated_envs = []
    cco_config = {}
    if _cco_is_configured() and _get_config():
        cco_config = _get_config()
        env_exists = any([environment.lower() == _env.get('name', '').lower() for _env in cco_config.get('environments') or []])
        if not env_exists:
            logs.info(f'Environment {environment} does not exists. Please run `cco ckan env add` with the same flags')
            return
        envs = cco_config.get('environments', [])

        for _env in envs:
            if _env.get('name').lower() == environment.lower():
                to_update = {k:v for k,v in kwargs.items() if v is not None}
                _env.update(to_update)
            updated_envs.append(_env)
        cco_config['environments'] = updated_envs
    _mkconfdir()
    _write_yaml(cco_config)


def set_environment(environment):
    updated_envs = []
    cco_config = {}
    if _cco_is_configured() and _get_config():
        cco_config = _get_config()
        env_exists = any([environment.lower() == _env.get('name', '').lower() for _env in cco_config.get('environments') or []])
        if not env_exists:
            logs.info(f'Environment {environment} does not exists. Please run `cco ckan env add` with the same flags')
            return
        envs = cco_config.get('environments', [])

        for _env in envs:
            _env['active'] = _env.get('name').lower() == environment.lower()
            if _env['active']:
                add_environment(environment, **_env)
            updated_envs.append(_env)

        cco_config['environments'] = updated_envs
    _mkconfdir()
    _write_yaml(cco_config)



def remove_environment(environment):
    updated_envs = []
    cco_config = {}
    if _cco_is_configured() and _get_config():
        cco_config = _get_config()
        env_exists = any([environment.lower() == _env.get('name', '').lower() for _env in cco_config.get('environments') or []])
        if not env_exists:
            logs.info(f'Environment {environment} does not exists. Please run `cco ckan env add {environment}`')
            return
        envs = cco_config.get('environments', [])

        for _env in envs:
            if _env.get('name').lower() == environment.lower():
                continue
            updated_envs.append(_env)
        cco_config['environments'] = updated_envs
    _mkconfdir()
    _write_yaml(cco_config)


def _cco_is_configured():
    return os.path.isfile(_get_config_file_name())


def _get_config():
    config_file = open(_get_config_file_name())
    return yaml.load(config_file)


def _get_config_file_name():
    return str(os.path.join(Path.home(),'.cco','cco.yaml'))


def _get_config_dir():
    return str(os.path.join(Path.home(),'.cco'))


def _write_yaml(data):
    _file = open(_get_config_file_name(), "w")
    yaml_config.yaml_dump(data, _file)
    _file.close()


def _read_yaml():
    with open(_get_config_file_name()) as _file:
        yaml_conf = yaml.load(_file, Loader=yaml.FullLoader)
    return yaml_conf


def _mkconfdir():
    dir = _get_config_dir()
    Path(f'{dir}').mkdir(parents=True, exist_ok=True)


def _run_az_command_or_login(command, func, *args):
    if not args[-1]:
        logs.subprocess_check_output(['az', 'login'])
        print("Logged in successfully.")
    try:
        logs.subprocess_check_call(command.split(' '))
    except AttributeError as e:
        print('You need to login to azure CLI. Please authenticate with browser.')
        func(*args[:-1], logged_in=False)



import click
import yaml

from . import manager


@click.group('providers')
def providers_group():
    """Manage configurable providers of centralized infrastructure services"""
    pass


@providers_group.command()
def list_providers():
    print(yaml.dump(manager.list_providers(), default_flow_style=False))



import yaml
import base64
import traceback
import datetime
import json
import logging
import subprocess
from ckan_cloud_operator import yaml_config
from ckan_cloud_operator import logs


def check_call(cmd, namespace='ckan-cloud', use_first_pod=False):
    cmd = _parse_call_cmd(cmd, namespace, use_first_pod)
    logs.subprocess_check_call(f'kubectl -n {namespace} {cmd}', shell=True)


def get_deployment_pod_name(deployment_name, namespace='ckan-cloud', use_first_pod=False, required_phase=None):
    deployment = get(f'deployment {deployment_name}', namespace=namespace, required=True)
    match_labels = deployment['spec']['selector']['matchLabels']
    pods = get_items_by_labels('pod', match_labels, required=True, namespace=namespace)
    if use_first_pod:
        assert len(pods) > 0
    else:
        assert len(pods) == 1
    if required_phase:
        assert pods[0]['status']['phase'].lower() == required_phase.lower()
    return pods[0]['metadata']['name']


def check_output(cmd, namespace='ckan-cloud'):
    return logs.subprocess_check_output(f'kubectl -n {namespace} {cmd}', shell=True)


def call(cmd, namespace='ckan-cloud'):
    return subprocess.call(f'kubectl -n {namespace} {cmd}', shell=True)


def getstatusoutput(cmd, namespace='ckan-cloud', use_first_pod=False):
    cmd = _parse_call_cmd(cmd, namespace, use_first_pod)
    return subprocess.getstatusoutput(f'kubectl -n {namespace} {cmd}')


def get(what, *args, required=True, namespace='ckan-cloud', get_cmd='get', **kwargs):
    extra_args = ' '.join(args)
    extra_kwargs = ' '.join([f'{k} {v}' for k, v in kwargs.items()])
    try:
        return yaml.load(
            logs.subprocess_check_output(
                f'kubectl -n {namespace} {get_cmd} {what} {extra_args} -o yaml {extra_kwargs}', shell=True,
            )
        )
    except subprocess.CalledProcessError:
        if required:
            raise
        else:
            return None


def edit(what, *edit_args, namespace='ckan-cloud', **edit_kwargs):
    extra_edit_args = ' '.join(edit_args)
    extra_edit_kwargs = ' '.join([f'{k} {v}' for k, v in edit_kwargs.items()])
    res = get(what, namespace=namespace, required=True)
    items = res.get('items', [res])
    assert len(items) > 0, f'no items found to edit for: {what}'
    for item in items:
        name = item['metadata']['name']
        kind = item['kind']
        subprocess.check_call(f'kubectl -n {namespace} edit {kind}/{name} {extra_edit_args} {extra_edit_kwargs}', shell=True)


def get_items_by_labels(resource_kind, labels, required=True, namespace='ckan-cloud'):
    if labels:
        label_selector = ','.join([f'{k}={v}' for k,v in labels.items()])
        label_args = f'-l {label_selector}'
    else:
        label_args = ''
    res = get(f'{resource_kind} {label_args}', required=required, namespace=namespace)
    return res['items'] if res else None


def edit_items_by_labels(resource_kind, labels, namespace='ckan-cloud'):
    label_selector = ','.join([f'{k}={v}' for k,v in labels.items()])
    edit(f'{resource_kind} -l {label_selector}', namespace=namespace)


def delete_items_by_labels(resource_kinds, labels, namespace='ckan-cloud'):
    label_selector = ','.join([f'{k}={v}' for k,v in labels.items()])
    resource_kinds = ','.join(resource_kinds)
    check_call(f'delete --ignore-not-found  -l {label_selector} {resource_kinds}', namespace=namespace)


def decode_secret(secret, attr=None, required=False):
    if not secret:
        if required:
            raise Exception()
        else:
            return {}
    elif attr:
        if required:
            return base64.b64decode(secret['data'][attr]).decode()
        else:
            value = secret.get('data', {}).get(attr)
            if value:
                return base64.b64decode(value)
            else:
                return None
    else:
        return {
            k: base64.b64decode(v).decode() if v else None
            for k, v
            in secret.get('data', {}).items()
        }


def update_secret(name, values, namespace='ckan-cloud', labels=None, dry_run=False):
    for k, v in values.items():
        v_type = type(v)
        assert v_type == str, f'Invalid type ({v_type}) for {k}: {v}'
    if not labels:
        labels = {}
    secret = get(f'secret {name}', required=False, namespace=namespace)
    labels = dict(secret.get('metadata', {}).get('labels', {}), **labels) if secret else labels
    data = decode_secret(secret, required=False)
    data.update(**values)
    apply({
        'apiVersion': 'v1',
        'kind': 'Secret',
        'metadata': {
            'name': name,
            'namespace': namespace,
            'labels': labels,
        },
        'type': 'Opaque',
        'data': {k: base64.b64encode(v.encode()).decode() for k, v in data.items() if v}
    }, dry_run=dry_run)
    return data


def update_configmap(name, values, namespace='ckan-cloud', labels=None, dry_run=False):
    for k, v in values.items():
        v_type = type(v)
        assert v_type == str, f'Invalid type ({v_type}) for {k}: {v}'
    configmap = get(f'configmap {name}', required=False, namespace=namespace)
    data = configmap['data'] if configmap else {}
    data.update(**values)
    apply(get_configmap(name, labels, data, namespace=namespace), dry_run=dry_run)
    return data


def get_item_detailed_status(item):
    kind = item['kind']
    name = item['metadata']['name']
    item_status = {"name": name, "created_at": item["metadata"]["creationTimestamp"], "true_status_last_transitions": {}}
    if kind in ["Deployment", "ReplicaSet"]:
        item_status["generation"] = item["metadata"]["generation"]
    for condition in item["status"].get("conditions", []):
        assert condition["type"] not in item_status["true_status_last_transitions"]
        if condition["status"] == "True":
            item_status["true_status_last_transitions"][condition["type"]] = condition["lastTransitionTime"]
        else:
            item_status.setdefault("errors", []).append({
                "kind": "failed_condition",
                "status": condition["status"],
                "reason": condition["reason"],
                "message": condition["message"],
                "last_transition": condition["lastTransitionTime"]
            })
    return item_status


def get_deployment_detailed_status(deployment, pod_label_selector, main_container_name, namespace='ckan-cloud'):
    status = get_item_detailed_status(deployment)
    ready = len(status.get('error', [])) == 0
    status['pods'] = []
    pods = get(f'pods -l {pod_label_selector}', namespace=namespace, required=False)
    if pods:
        for pod in pods['items']:
            pod_status = get_item_detailed_status(pod)
            pod_status['other-containers'] = []
            for container in pod['spec']['containers']:
                container_name = container['name']
                container_status = {'name': container_name}
                status_code, output = subprocess.getstatusoutput(
                    f'kubectl -n {namespace} logs {pod["metadata"]["name"]} -c {main_container_name} --tail 5',
                )
                if status_code == 0:
                    container_status['logs'] = output
                else:
                    if container_name == main_container_name: ready = False
                    container_status['logs'] = ''
                container_status['image'] = container['image']
                if container_name == main_container_name:
                    pod_status['main-container'] = container_status
                else:
                    pod_status['other-containers'].append(container_status)
            status['pods'].append(pod_status)
    return dict(status, ready=ready, namespace=deployment['metadata']['namespace'])


def create(resource, is_yaml=False):
    if is_yaml: resource = yaml.load(resource)
    try:
        logs.subprocess_run('kubectl create -f -', input=yaml.dump(resource).encode())
    except:
        logging.exception('Failed to create resource\n%s', yaml.dump(resource, default_flow_style=False))
        raise



def apply(resource, is_yaml=False, reconcile=False, dry_run=False):
    if is_yaml: resource = yaml.load(resource)
    cmd = 'auth reconcile' if reconcile else 'apply'
    args = []
    if dry_run:
        args.append('--dry-run')
    args = " ".join(args)
    try:
        logs.subprocess_run(
            f'kubectl {cmd} {args} -f -',
            input=yaml.dump(resource).encode()
        )
    except:
        logging.exception('Failed to apply resource\n%s', yaml.dump(resource, default_flow_style=False))
        raise
    if dry_run:
        print(yaml.dump(resource, default_flow_style=False))


def install_crd(plural, singular, kind):
    crd = get(f'crd {plural}.stable.viderum.com', required=False)
    version = 'v1'
    if crd:
        assert crd['spec']['versions'][0]['name'] == version
        print(f'{kind} custom resource definitions are up-to-date')
    else:
        print(f'Creating {kind} {version} custom resource definition')
        crd = {'apiVersion': 'apiextensions.k8s.io/v1beta1',
               'kind': 'CustomResourceDefinition',
               'metadata': {
                   'name': f'{plural}.stable.viderum.com'
               },
               'spec': {
                   'versions': [{
                       'name': version,
                       'served': True,
                       'storage': True
                   }],
                   'group': 'stable.viderum.com',
                   'scope': 'Namespaced',
                   'names': {
                       'plural': plural,
                       'singular': singular,
                       'kind': kind
                   }
               }}
        logs.subprocess_run('kubectl create -f -', input=yaml.dump(crd).encode())


def get_resource(api_version, kind, name, labels, namespace='ckan-cloud', **kwargs):
    resource = {
        'apiVersion': api_version,
        'kind': kind,
        'metadata': {
            'name': name,
            'namespace': namespace,
            'labels': labels,
            'annotations': {},
        },
    }
    resource.update(**kwargs)
    add_operator_timestamp_annotation(resource['metadata'])
    return resource


def get_configmap(name, labels, data, namespace='ckan-cloud'):
    configmap = get_resource('v1', 'ConfigMap', name, labels, namespace)
    return dict(configmap, data=data)


def get_persistent_volume_claim(name, labels, spec, namespace='ckan-cloud'):
    pvc = get_resource('v1', 'PersistentVolumeClaim', name, labels, namespace)
    return dict(pvc, spec=spec)


def get_deployment(name, labels, spec, namespace='ckan-cloud', with_timestamp=True):
    deployment = get_resource('apps/v1', 'Deployment', name, labels, namespace)
    deployment = dict(deployment, spec=spec)
    if with_timestamp:
        add_operator_timestamp_annotation(deployment['spec']['template']['metadata'])
    return deployment


def get_service(name, labels, ports, selector, namespace='ckan-cloud'):
    service = get_resource('v1', 'Service', name, labels, namespace)
    service['spec'] = {
        'ports': [
            {'name': str(port), 'port': int(port)}
            for port in ports
        ],
        'selector': selector
    }
    return service


def now():
    return datetime.datetime.now().isoformat()


def timestamp():
    return str(int(datetime.datetime.timestamp(datetime.datetime.now())))


def add_operator_timestamp_annotation(metadata):
    metadata.setdefault('annotations', {})['ckan-cloud/operator-timestamp'] = now()


def remove_finalizers(resource_kind, resource_name, ignore_not_found=False):
    if ignore_not_found and not get(f'{resource_kind} {resource_name}', required=False):
        return True
    else:
        return call(f'patch {resource_kind} {resource_name} -p \'{{"metadata":{{"finalizers":[]}}}}\' --type=merge') == 0


def remove_resource_and_dependencies(resource_kind, resource_name, related_kinds, label_selector):
    kinds = ','.join(related_kinds)
    return all([
        call(f'delete --ignore-not-found -l {label_selector} {kinds}') == 0,
        call(f'delete --ignore-not-found {resource_kind}/{resource_name}') == 0,
        remove_finalizers(resource_kind, resource_name, ignore_not_found=True)
    ])


__NONE__ = object


class BaseAnnotations(object):
    """Base class for managing annotations related to a custom resource"""

    @property
    def FLAGS(self):
        """Boolean flags which are saved as annotations on the resource"""
        return ['forceCreateAnnotations']

    @property
    def STATUSES(self):
        """Predefined statuses which are saved as annotations on the resource"""
        return {}

    @property
    def SECRET_ANNOTATIONS(self):
        """Sensitive details which are saved in a secret related to the resource"""
        return []

    @property
    def JSON_ANNOTATION_PREFIXES(self):
        """flexible annotations encoded to json and permitted using key prefixes"""
        return []

    @property
    def RESOURCE_KIND(self):
        raise NotImplementedError()

    def get_secret_labels(self):
        return {'ckan-cloud/annotations-secret': self.resource_id}

    def __init__(self, resource_id, resource_values=None, override_flags=None, persist_overrides=False):
        """Initializes an annotations object, optionally specifying previously fetched resource values"""
        self.resource_id = resource_id
        self.resource_values = resource_values
        self.resource_kind = self.RESOURCE_KIND.lower()
        self._override_flags = override_flags
        if self._override_flags:
            for flag in self._override_flags:
                assert flag in self.FLAGS, f'unknown flag: {flag}'
            if persist_overrides and len(self._override_flags) > 0:
                self._annotate(*['{}={}'.format(k, 'true' if v else 'false')
                                 for k, v in self._override_flags.items()], overwrite=True)

    def set_status(self, key, status):
        assert status in self.STATUSES[key], f'unknown status/key: {status}/{key}'
        self._annotate(f'{key}-{status}=true')

    def get_status(self, key, status):
        assert key in self.STATUSES.keys(), f'unknown status key: {key}'
        return bool(self._get_annotation(f'{key}-{status}'))

    def update_status(self, key, status, update_func, force_update=False):
        if self.get_status(key, status):
            if force_update:
                update_func()
                return True
            else:
                return False
        else:
            try:
                update_func()
            except:
                if self.get_flag('forceCreateAnnotations'):
                    traceback.print_exc()
                else:
                    raise
            self.set_status(key, status)
            return True

    def set_flag(self, flag):
        self._annotate(f'{flag}=true', overwrite=True)

    def set_flags(self, *flags):
        self._annotate(*[f'{flag}=true' for flag in flags], overwrite=True)

    def update_flag(self, flag, update_func, force_update=False):
        if self.get_flag(flag):
            if force_update:
                update_func()
                return True
            else:
                return False
        else:
            update_func()
            self.set_flag(flag)
            return True

    def get_flag(self, flag):
        """flags  are boolean fields which can be added manually to modify operator functionality

        Flags can be set using kubectl annotate on the resource, for example:

        kubectl -n ckan-cloud annotate DeisCkanInstance/atea32 ckan-cloud/forceCreateAnnotations=true
        """
        assert flag in self.FLAGS, f'unknown flag: {flag}'
        if self._override_flags and flag in self._override_flags:
            return self._override_flags[flag]
        else:
            value = self._get_annotation(flag)
            if type(value) == str:
                return False if value.lower().strip() in ['false', '0', 'no', ''] else bool(value)
            else:
                return bool(value)

    def get(self):
        data = {k.replace('ckan-cloud/', ''): v for k, v in self.resource_values['metadata']['annotations'].items()
                if k.startswith('ckan-cloud/')}
        data['ready'] = True
        return data

    def set_secrets(self, key_values):
        for key in key_values:
            assert key in self.SECRET_ANNOTATIONS, 'unknown secret annotation: {key}'
        secret = getattr(self, '_secret', None)
        cur_data = secret.get('data', {}) if secret and secret != __NONE__ else {}
        secret = get(f'secret {self.resource_kind}-{self.resource_id}-annotations', required=False)
        if not secret:
            secret = {'data': {}}
        secret['data'].update(**cur_data)
        for key, value in key_values.items():
            secret['data'][key] = base64.b64encode(value.encode()).decode()
        secret = {
            'apiVersion': 'v1',
            'kind': 'Secret',
            'metadata': {
                'name': f'{self.resource_kind}-{self.resource_id}-annotations',
                'namespace': 'ckan-cloud',
                'labels': self.get_secret_labels()
            },
            'type': 'Opaque',
            'data': secret['data']
        }
        logs.subprocess_run(f'kubectl apply -f -', input=yaml.dump(secret).encode())
        self._secret = secret

    def set_secret(self, key, value):
        self.set_secrets({key: value})

    def get_secret(self, key, default=None):
        assert key in self.SECRET_ANNOTATIONS, f'unknown secret annotation: {key}'
        secret = getattr(self, '_secret', None)
        if not secret:
            secret = get(f'secret {self.resource_kind}-{self.resource_id}-annotations', required=False)
            if not secret:
                secret = __NONE__
            self._secret = secret
        if secret and secret != __NONE__:
            value = secret.get('data', {}).get(key, None)
            return base64.b64decode(value).decode() if value else default
        else:
            return default

    def get_pod_env_spec_from_secret(self, env_name, value_key):
        return {
            'name': env_name,
            'valueFrom': {
                'secretKeyRef': {
                    'name': f'{self.resource_kind}-{self.resource_id}-annotations',
                    'key': value_key
                }
            }
        }

    def json_annotate(self, key, value, overwrite=True):
        ans = []
        assert any([key.startswith(prefix) for prefix in self.JSON_ANNOTATION_PREFIXES]), f'invalid json annotation key: {key}'
        value = json.dumps(value)
        ans.append(f"{key}='{value}'")
        self._annotate(*ans, overwrite=overwrite)

    def get_json_annotation(self, key):
        return json.loads(self._get_annotation(key))

    def _annotate(self, *annotations, overwrite=False):
        cmd = f'kubectl -n ckan-cloud annotate {self.resource_kind} {self.resource_id}'
        for annotation in annotations:
            cmd += f' ckan-cloud/{annotation}'
        if overwrite:
            cmd += ' --overwrite'
        logs.subprocess_check_call(cmd, shell=True)

    def _get_annotation(self, annotation, default=None):
        return self.resource_values['metadata'].get('annotations', {}).get(f'ckan-cloud/{annotation}', default)


def _parse_call_cmd(cmd, namespace, use_first_pod):
    args = []
    for arg in cmd.split(' '):
        splitarg = arg.split(':')
        if splitarg[0] == 'deployment-pod':
            _, deployment_namespace, deployment_name = splitarg
            if not deployment_namespace:
                deployment_namespace = namespace
            arg = get_deployment_pod_name(deployment_name, namespace=deployment_namespace, use_first_pod=use_first_pod)
        args.append(arg)
    return ' '.join(args)
